"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

import cvxpy.lin_ops.lin_utils as lu
from cvxpy import settings as s
from cvxpy.expressions.expression import Expression
from cvxpy.expressions.leaf import Leaf
from cvxpy.utilities import scopes


def is_param_affine(expr) -> bool:
    """Returns true if expression is parameters-affine (and variable-free)"""
    with scopes.dpp_scope():
        return not expr.variables() and expr.is_affine()


def is_param_free(expr) -> bool:
    """Returns true if expression is not parametrized."""
    return not expr.parameters()


class Parameter(Leaf):
    """Parameters in optimization problems.

    Parameters are constant expressions whose value may be specified
    after problem creation. The only way to modify a problem after its
    creation is through parameters. For example, you might choose to declare
    the hyper-parameters of a machine learning model to be Parameter objects;
    more generally, Parameters are useful for computing trade-off curves.
    """
    PARAM_COUNT = 0

    def __init__(
        self, shape: int | tuple[int, ...] = (), name: str | None = None, value=None,
        id=None, **kwargs
    ) -> None:
        bounds = kwargs.get('bounds')
        if bounds is not None:
            for b in bounds:
                if isinstance(b, Expression):
                    raise ValueError(
                        "Parameter bounds must be numeric, not CVXPY "
                        "Expressions. Parametric bounds are only "
                        "supported on Variables."
                    )
        if id is None:
            self.id = lu.get_id()
        else:
            self.id = id
        if name is None:
            self._name = f"{s.PARAM_PREFIX}{self.id}"
        else:
            self._name = name
        # Initialize with value if provided.
        self._value = None
        self.delta = None
        self.gradient = None
        super(Parameter, self).__init__(shape, value, **kwargs)
        self._is_constant = True

    def get_data(self):
        """Returns info needed to reconstruct the expression besides the args.
        """
        return [self.shape, self._name, self.value, self.id, self.attributes]

    def name(self) -> str:
        return self._name

    def is_constant(self) -> bool:
        if scopes.dpp_scope_active():
            return False
        return True

    @property
    def grad(self):
        """Gives the (sub/super)gradient of the expression w.r.t. each variable.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Returns:
            A map of variable to SciPy CSC sparse matrix or None.
        """
        return {}

    def parameters(self) -> list[Parameter]:
        """Returns itself as a parameter.
        """
        return [self]

    def canonicalize(self):
        """Returns the graph implementation of the object.

        Returns:
            A tuple of (affine expression, [constraints]).
        """
        obj = lu.create_param(self.shape, self.id)
        return (obj, [])

    def __repr__(self) -> str:
        """String to recreate the object.
        """
        attr_str = self._get_attr_str()
        if len(attr_str) > 0:
            return "Parameter(%s%s)" % (self.shape, attr_str)
        else:
            return "Parameter(%s)" % (self.shape,)



"""Tests for cp.nlp namespace."""
import cvxpy as cp
from cvxpy.atoms.elementwise.hyperbolic import tanh
from cvxpy.atoms.elementwise.normcdf import normcdf
from cvxpy.atoms.elementwise.trig import cos, sin


class TestNLPNamespace:
    def test_nlp_namespace_accessible(self):
        """Test that cp.nlp submodule is accessible."""
        assert hasattr(cp, 'nlp')

    def test_trig_atoms(self):
        x = cp.Variable()
        assert cp.nlp.sin(x) is not None
        assert cp.nlp.cos(x) is not None
        assert cp.nlp.tan(x) is not None
        assert cp.nlp.normcdf(x) is not None

    def test_hyperbolic_atoms(self):
        x = cp.Variable()
        assert cp.nlp.sinh(x) is not None
        assert cp.nlp.tanh(x) is not None
        assert cp.nlp.asinh(x) is not None
        assert cp.nlp.atanh(x) is not None

    def test_atoms_not_in_top_level(self):
        """NLP atoms should only be accessible via cp.nlp, not cp directly."""
        assert not hasattr(cp, 'sin')
        assert not hasattr(cp, 'cos')
        assert not hasattr(cp, 'tan')
        assert not hasattr(cp, 'sinh')
        assert not hasattr(cp, 'tanh')
        assert not hasattr(cp, 'asinh')
        assert not hasattr(cp, 'atanh')
        assert not hasattr(cp, 'normcdf')

    def test_atoms_same_class_as_direct_import(self):
        """cp.nlp.sin should be the same class as a direct import."""
        assert cp.nlp.sin is sin
        assert cp.nlp.cos is cos
        assert cp.nlp.tanh is tanh
        assert cp.nlp.normcdf is normcdf



"""
Copyright 2025 CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np

from cvxpy.atoms.affine.sum import Sum
from cvxpy.atoms.elementwise.power import power
from cvxpy.expressions.variable import Variable
from cvxpy.reductions.dnlp2smooth.canonicalizers.power_canon import power_canon

MIN_INIT = 1

def quad_over_lin_canon(expr, args):
    """
    Canonicalize a quadratic over linear expression.
    If the denominator is constant, we can use the power canonicalizer.
    Otherwise, we introduce new variables for the numerator and denominator.
    """
    if args[1].is_constant():
        expr = power(args[0], 2)
        var, constr = power_canon(expr, expr.args)
        summation = Sum(var)
        return 1/args[1].value * summation, constr
    else:
        t1 = args[0]
        t2 = args[1]
        constraints = []
        if not isinstance(t1, Variable):
            t1 = Variable(t1.shape)
            constraints += [t1 == args[0]]
            t1.value = args[0].value
        # always introduce a new variable for the denominator
        # so that we can initialize it to 1 (point in domain)
        t2 = Variable(t2.shape, nonneg=True)
        constraints += [t2 == args[1]]

        if args[1].value is not None:
            t2.value = np.maximum(args[1].value, MIN_INIT)

        return expr.copy([t1, t2]), constraints



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np

from cvxpy import error
from cvxpy.problems.objective import Maximize
from cvxpy.reductions.cvx_attr2constr import CvxAttr2Constr
from cvxpy.reductions.dnlp2smooth.dnlp2smooth import Dnlp2Smooth
from cvxpy.reductions.flip_objective import FlipObjective
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS, NLP_SOLVER_VARIANTS, SOLVER_MAP_NLP
from cvxpy.reductions.solvers.solving_chain import SolvingChain


def _build_nlp_chain(problem, solver, kwargs):
    """Build the NLP reduction chain and return (SolvingChain, kwargs).

    Solver selection may mutate kwargs (e.g., Knitro algorithm, Uno preset).
    """
    # Resolve the solver instance.
    if solver is None:
        # Pick first installed NLP solver in preference order.
        for name, inst in SOLVER_MAP_NLP.items():
            if name in INSTALLED_SOLVERS:
                solver_instance = inst
                break
        else:
            raise error.SolverError(
                "No NLP solver is installed. Install one of: %s"
                % ", ".join(SOLVER_MAP_NLP)
            )
    elif solver in SOLVER_MAP_NLP:
        solver_instance = SOLVER_MAP_NLP[solver]
    elif solver.lower() in NLP_SOLVER_VARIANTS:
        base_name, variant_kwargs = NLP_SOLVER_VARIANTS[solver.lower()]
        kwargs.update(variant_kwargs)
        solver_instance = SOLVER_MAP_NLP[base_name]
    else:
        raise error.SolverError(
            "Solver %s is not supported for NLP problems." % solver
        )

    # Build the reduction chain.
    if type(problem.objective) == Maximize:
        reductions = [FlipObjective()]
    else:
        reductions = []
    reductions += [
        CvxAttr2Constr(reduce_bounds=not solver_instance.BOUNDED_VARIABLES),
        Dnlp2Smooth(),
        solver_instance,
    ]

    return SolvingChain(reductions=reductions), kwargs


def _set_nlp_initial_point(problem):
    """Construct an initial point for variables without a user-specified value.

    Uses get_bounds() which incorporates sign attributes (nonneg, nonpos, etc.).
    If both lb and ub are finite, initialize to their midpoint. If only one is
    finite, initialize one unit from the bound. Otherwise, initialize to zero.
    """
    for var in problem.variables():
        if var.value is not None:
            continue

        lb, ub = var.get_bounds()

        lb_finite = np.isfinite(lb)
        ub_finite = np.isfinite(ub)

        init = np.zeros(var.shape)
        both = lb_finite & ub_finite
        lb_only = lb_finite & ~ub_finite
        ub_only = ~lb_finite & ub_finite
        init[both] = 0.5 * (lb[both] + ub[both])
        init[lb_only] = lb[lb_only] + 1.0
        init[ub_only] = ub[ub_only] - 1.0

        var.save_value(init)


def _set_random_nlp_initial_point(problem, run, user_initials):
    """Generate a random initial point for DNLP problems.

    A variable is initialized randomly if:
    1. 'sample_bounds' is set for that variable.
    2. The initial value specified by the user is None, 'sample_bounds' is not
       set, but the variable has both finite lower and upper bounds.

    Parameters
    ----------
    problem : Problem
    run : int
        Current run index (0-based).
    user_initials : dict
        On run 0, will be populated with user-specified initial values.
        On subsequent runs, used to restore user-specified values.
    """
    # Store user-specified initial values on the first run
    if run == 0:
        user_initials.clear()
        for var in problem.variables():
            if var.sample_bounds is not None:
                user_initials[var.id] = None
            else:
                user_initials[var.id] = var.value

    for var in problem.variables():
        # Skip variables with user-specified initial value
        # (note that any variable with sample bounds set will have
        #  user_initials[var.id] == None)
        if user_initials[var.id] is not None:
            # Reset to user-specified initial value from last solve
            var.value = user_initials[var.id]
            continue
        else:
            # Reset to None from last solve
            var.value = None

        # Determine effective sample bounds: use explicit sample_bounds if set,
        # otherwise fall back to variable bounds.
        sb = var.sample_bounds
        if sb is None:
            sb = var.get_bounds()

        # Sample initial value if effective sample bounds are available. Otherwise
        # raise an error.
        if sb is not None:
            low, high = sb
            if not np.all(np.isfinite(low)) or not np.all(np.isfinite(high)):
                raise ValueError(
                    "Variable %s has non-finite sample_bounds. Cannot generate"
                    " random initial point. Either add sample bounds or set the value. "
                    " You can add sample bounds via var.sample_bounds = (low, high)."
                    % (var.name())
                )

            initial_val = np.random.uniform(low=low, high=high, size=var.shape)
            var.save_value(initial_val)


def solve_nlp(problem, solver, warm_start, verbose, **kwargs):
    """Solve an NLP problem using the DNLP reduction chain.

    Parameters
    ----------
    problem : Problem
        A DNLP-valid problem.
    solver : str or None
        Solver name (e.g., 'IPOPT', 'knitro_sqp').
    warm_start : bool
        Whether to warm-start the solver.
    verbose : bool
        Whether to print solver output.
    **kwargs
        Additional solver options, including 'best_of'.

    Returns
    -------
    float
        The optimal problem value.
    """
    nlp_chain, kwargs = _build_nlp_chain(problem, solver, kwargs)

    # Reuse cached Oracles across solve() calls when problem has parameters.
    # Parameter updates happen inside solve_via_data when reusing.
    # Also initialize an empty cache for best_of.
    solver_cache = problem._solver_cache.get('NLP')
    if solver_cache is None and (problem.parameters() or "best_of" in kwargs):
        solver_cache = {}
        problem._solver_cache['NLP'] = solver_cache

    if "best_of" not in kwargs:
        _set_nlp_initial_point(problem)
        canon_problem, inverse_data = nlp_chain.apply(problem=problem)
        solution = nlp_chain.solver.solve_via_data(
            canon_problem, warm_start, verbose, solver_opts=kwargs,
            solver_cache=solver_cache)
        problem.unpack_results(solution, nlp_chain, inverse_data)
        return problem.value

    best_of = kwargs.pop("best_of")
    if not isinstance(best_of, int) or best_of < 1:
        raise ValueError("best_of must be a positive integer.")

    # Best-of-N solve
    best_obj, best_solution = float("inf"), None
    all_objs = np.zeros(shape=(best_of,))
    user_initials = {}

    for run in range(best_of):
        _set_random_nlp_initial_point(problem, run, user_initials)
        canon_problem, inverse_data = nlp_chain.apply(problem=problem)
        solution = nlp_chain.solver.solve_via_data(canon_problem, warm_start,
                                                   verbose, solver_opts=kwargs,
                                                   solver_cache=solver_cache)

        # unpack to get the objective value in the original problem space.
        # (+inf for infeasible runs, -inf for unbounded runs)
        problem.unpack_results(solution, nlp_chain, inverse_data)
        obj_value = problem.value

        all_objs[run] = obj_value

        # always set best_solution with the first run so that even an
        # all-infeasible best_of has a solution to unpack at the end (its
        # INFEASIBLE status then propagates through unpack_results).
        if best_solution is None or obj_value < best_obj:
            best_obj = obj_value
            best_solution = solution

        if verbose:
            print("Run %d/%d: obj = %.6e | best so far = %.6e"
                  % (run + 1, best_of, obj_value, best_obj))
            print("-" * 60)

    # Unpack best solution
    if type(problem.objective) == Maximize:
        all_objs = -all_objs

    # Propagate all objective values to the user
    best_solution['all_objs_from_best_of'] = all_objs
    problem.unpack_results(best_solution, nlp_chain, inverse_data)
    return problem.value



"""
Copyright 2017 Robin Verschueren

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import abc

from cvxpy import settings as s
from cvxpy.reductions.cone2cone.approx import APPROX_CONE_CONVERSIONS
from cvxpy.reductions.cone2cone.exact import EXACT_CONE_CONVERSIONS
from cvxpy.reductions.reduction import Reduction
from cvxpy.reductions.solvers.solver_inverse_data import SolverInverseData


def expand_cones(cones, supported):
    """Expand unsupported cones via exact and approximate conversions.

    Parameters
    ----------
    cones : set[type]
        Mutable set of cone constraint types (modified in place).
    supported : frozenset[type]
        Cones the target solver natively supports.

    Returns
    -------
    (cones, exact_targets, approx_targets) where *cones* is the
    (mutated) input set and the targets are the cones that were expanded.
    """
    # Compute which unsupported cones can transitively reach a
    # supported cone via the EXACT_CONE_CONVERSIONS DAG.
    reachable = set()
    changed = True
    while changed:
        changed = False
        for src, tgts in EXACT_CONE_CONVERSIONS.items():
            if src not in reachable and tgts & (supported | reachable):
                reachable.add(src)
                changed = True

    exact_targets = set()
    while True:
        new_targets = (cones & reachable) - supported
        if not new_targets:
            break
        exact_targets |= new_targets
        for co in new_targets:
            cones.discard(co)
            cones.update(EXACT_CONE_CONVERSIONS[co])

    approx_targets = (cones & APPROX_CONE_CONVERSIONS.keys()) - supported
    for co in approx_targets:
        cones.discard(co)
        cones.update(APPROX_CONE_CONVERSIONS[co])

    return cones, exact_targets, approx_targets


class Solver(Reduction):
    """Generic interface for a solver that uses reduction semantics
    """

    DIMS = "dims"
    # ^ The key that maps to "ConeDims" in the data returned by apply().
    #
    #   There are separate ConeDims classes for cone programs vs QPs.
    #   See cone_matrix_stuffing.py and qp_matrix_stuffing.py for details.

    # Solver capabilities.
    MIP_CAPABLE = False
    BOUNDED_VARIABLES = False
    SOC_DIM3_ONLY = False

    # PSD constraint format. Overridden by solvers that support PSD constraints.
    # PSD_TRIANGLE_KIND: which triangle the solver expects (TriangleKind.LOWER
    # or TriangleKind.UPPER). None means no PSD support via this mechanism.
    # PSD_SQRT2_SCALING: whether off-diagonal entries are scaled by sqrt(2).
    PSD_TRIANGLE_KIND = None
    PSD_SQRT2_SCALING = None

    # Constraint support (overridden by ConicSolver and QpSolver).
    SUPPORTED_CONSTRAINTS = []
    REQUIRES_CONSTR = False

    # Keys for inverse data.
    VAR_ID = 'var_id'
    DUAL_VAR_ID = 'dual_var_id'
    EQ_CONSTR = 'eq_constr'
    NEQ_CONSTR = 'other_constr'

    @abc.abstractmethod
    def name(self):
        """The name of the solver.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def import_solver(self):
        """Imports the solver.
        """
        raise NotImplementedError()

    def is_installed(self) -> bool:
        """Is the solver installed?
        """
        try:
            self.import_solver()
            return True
        except Exception as e:
            if not isinstance(e, ModuleNotFoundError):
                s.LOGGER.warning(
                    f"Encountered unexpected exception importing solver {self.name()}:\n"
                    + repr(e)
                )
            return False

    @abc.abstractmethod
    def solve_via_data(self, data, warm_start: bool, verbose: bool, solver_opts, solver_cache=None):
        """Solve a problem represented by data returned from apply.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def cite(self, data):
        """Return bibtex citation for the solver.
        """
        raise NotImplementedError()

    def supports_quad_obj(self) -> bool:
        """Whether the solver supports quadratic objectives."""
        return False

    def can_solve(self, problem_form) -> bool:
        """Check if this solver can handle a problem with the given structure.

        Parameters
        ----------
        problem_form : ProblemForm
            Pre-canonicalization structural analysis of the problem.
        """
        if problem_form.is_mixed_integer() and not self.MIP_CAPABLE:
            return False

        if problem_form.is_mixed_integer():
            supported = frozenset(
                getattr(self, 'MI_SUPPORTED_CONSTRAINTS', self.SUPPORTED_CONSTRAINTS)
            )
        else:
            supported = frozenset(self.SUPPORTED_CONSTRAINTS)

        quad_obj = self.supports_quad_obj() and problem_form.has_quadratic_objective()
        cones = problem_form.cones(quad_obj=quad_obj).copy()
        expand_cones(cones, supported)

        if not problem_form.has_constraints() and self.REQUIRES_CONSTR:
            return False

        return cones.issubset(supported)

    def solve(self, problem, warm_start: bool, verbose: bool, solver_opts):
        """Solve the problem and return a Solution object.
        """
        data, inv_data = self.apply(problem)
        solution = self.solve_via_data(data, warm_start, verbose, solver_opts)
        inverse_data = SolverInverseData(inv_data, solver_instance=self, solver_options=solver_opts)
        return self.invert(solution, inverse_data)



"""
Tests for DGP (Disciplined Geometric Programming) with DPP (Disciplined Parametrized Programming).

This verifies the fix for issue #3004: DGP problems can now have get_problem_data(gp=True)
called without all parameters having values. This is useful when the problem is DPP.
"""
import numpy as np
import pytest

import cvxpy as cp
from cvxpy.error import ParameterError
from cvxpy.tests.base_test import BaseTest

SOLVER = cp.CLARABEL


class TestDgpDpp(BaseTest):
    """Tests for DGP + DPP integration focusing on get_problem_data core functionality."""

    def test_get_problem_data_without_param_values(self) -> None:
        """get_problem_data(gp=True) works with uninitialized parameters.

        This was the original failure case in issue #3004. Previously, DGP
        canonicalization eagerly evaluated np.log(param.value), failing when
        param.value was None.
        """
        alpha = cp.Parameter(pos=True)  # No value set
        x = cp.Variable(pos=True)
        problem = cp.Problem(cp.Minimize(x), [x >= alpha])

        # Should not raise - canonicalization is pure structural transformation
        data, _, _ = problem.get_problem_data(SOLVER, gp=True)
        assert data is not None

        # Now set value and solve using the cached canonicalization
        alpha.value = 2.0
        problem.solve(SOLVER, gp=True, enforce_dpp=True)
        self.assertAlmostEqual(x.value, 2.0)

    def test_fast_path_with_changing_params(self) -> None:
        """DGP DPP fast path correctly updates parameters across solves.

        Tests that log-transformed parameter values are properly updated
        in the cached param_prog on each solve.
        """
        alpha = cp.Parameter(pos=True)
        x = cp.Variable(pos=True)
        problem = cp.Problem(cp.Minimize(x), [x >= alpha])

        # First solve
        alpha.value = 1.0
        problem.solve(SOLVER, gp=True, enforce_dpp=True)
        self.assertAlmostEqual(x.value, 1.0, places=4)

        # Second solve with different parameter value
        alpha.value = 2.0
        problem.solve(SOLVER, gp=True, enforce_dpp=True)
        self.assertAlmostEqual(x.value, 2.0, places=4)

        # Third solve
        alpha.value = 0.5
        problem.solve(SOLVER, gp=True, enforce_dpp=True)
        self.assertAlmostEqual(x.value, 0.5, places=4)

    def test_solve_without_param_value_raises_error(self) -> None:
        """Attempting to solve without setting parameter values raises ParameterError."""
        alpha = cp.Parameter(pos=True)
        x = cp.Variable(pos=True)
        problem = cp.Problem(cp.Minimize(x), [x >= alpha])

        # get_problem_data should work without values
        problem.get_problem_data(SOLVER, gp=True)

        # But solve should raise if values not set
        with pytest.raises(ParameterError, match="must have.*value.*before solving"):
            problem.solve(SOLVER, gp=True, enforce_dpp=True)

    def test_non_dpp_mode_with_ignore_dpp_flag(self) -> None:
        """DGP problems work correctly with ignore_dpp=True (non-DPP mode).

        This tests backward compatibility: even though the DGP canonicalization
        now supports DPP, forcing non-DPP treatment with ignore_dpp=True should
        still work correctly when parameters have values.
        """
        alpha = cp.Parameter(pos=True, value=1.0)
        x = cp.Variable(pos=True)
        problem = cp.Problem(cp.Minimize(x), [x >= alpha])

        # Problem is DPP-compatible but we force non-DPP treatment
        assert problem.is_dpp('dgp')
        problem.solve(SOLVER, gp=True, ignore_dpp=True)
        self.assertAlmostEqual(x.value, 1.0, places=4)

        # Update parameter and solve again with ignore_dpp
        alpha.value = 3.0
        problem.solve(SOLVER, gp=True, ignore_dpp=True)
        self.assertAlmostEqual(x.value, 3.0, places=4)


class TestDgpVariableBounds(BaseTest):
    """Tests for DGP variable bounds (numeric and parametric)."""

    def test_parametric_bounds(self) -> None:
        lb = cp.Parameter(pos=True, value=2.0)
        ub = cp.Parameter(pos=True, value=5.0)
        x = cp.Variable(pos=True, bounds=[lb, ub])

        cp.Problem(cp.Minimize(x)).solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 2.0, places=4)

        cp.Problem(cp.Maximize(x)).solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 5.0, places=4)

    def test_one_sided_parametric_bounds(self) -> None:
        p = cp.Parameter(pos=True, value=0.5)
        x = cp.Variable(pos=True, bounds=[p, None])
        cp.Problem(cp.Minimize(x), [x <= 10.0]).solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 0.5, places=4)

        q = cp.Parameter(pos=True, value=3.0)
        y = cp.Variable(pos=True, bounds=[None, q])
        cp.Problem(cp.Maximize(y)).solve(SOLVER, gp=True)
        self.assertAlmostEqual(y.value, 3.0, places=4)

    def test_mixed_numeric_and_parametric(self) -> None:
        ub = cp.Parameter(pos=True, value=4.0)
        x = cp.Variable(pos=True, bounds=[0.5, ub])
        cp.Problem(cp.Minimize(x)).solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 0.5, places=4)

    def test_re_solve(self) -> None:
        lb = cp.Parameter(pos=True, value=2.0)
        ub = cp.Parameter(pos=True, value=10.0)
        x = cp.Variable(pos=True, bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))

        prob.solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 2.0, places=4)

        lb.value = 3.0
        prob.solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 3.0, places=4)

    def test_is_dpp(self) -> None:
        p = cp.Parameter(pos=True)
        x = cp.Variable(pos=True, bounds=[p, None])
        assert x.is_dpp('dgp')
        assert cp.Problem(cp.Minimize(x), [x <= 10.0]).is_dpp('dgp')

        # Product of params: DPP in DGP (monomial), not in DCP.
        p2 = cp.Parameter(pos=True)
        y = cp.Variable(pos=True, bounds=[p * p2, None])
        assert y.is_dpp('dgp')
        assert not y.is_dpp('dcp')

    def test_enforce_dpp(self) -> None:
        lb = cp.Parameter(pos=True, value=2.0)
        x = cp.Variable(pos=True, bounds=[lb, None])
        prob = cp.Problem(cp.Minimize(x), [x <= 100.0])

        prob.solve(SOLVER, gp=True, enforce_dpp=True)
        self.assertAlmostEqual(x.value, 2.0, places=4)

        lb.value = 7.0
        prob.solve(SOLVER, gp=True, enforce_dpp=True)
        self.assertAlmostEqual(x.value, 7.0, places=4)

    def test_get_problem_data_without_param_values(self) -> None:
        p = cp.Parameter(pos=True)
        x = cp.Variable(pos=True, bounds=[p, None])
        data, _, _ = cp.Problem(cp.Minimize(x), [x <= 10.0]).get_problem_data(
            SOLVER, gp=True)
        assert data is not None

    def test_vector_and_matrix(self) -> None:
        lb = cp.Parameter(pos=True, value=2.0)
        x = cp.Variable(3, pos=True, bounds=[lb, None])
        cp.Problem(cp.Minimize(cp.sum(x)), [x <= 100.0]).solve(SOLVER, gp=True)
        np.testing.assert_allclose(x.value, 2.0 * np.ones(3), atol=1e-4)

        X = cp.Variable((2, 3), pos=True, bounds=[lb, None])
        cp.Problem(cp.Minimize(cp.sum(X)), [X <= 100.0]).solve(SOLVER, gp=True)
        np.testing.assert_allclose(X.value, 2.0 * np.ones((2, 3)), atol=1e-4)

    def test_equal_parametric_bounds(self) -> None:
        p = cp.Parameter(pos=True, value=3.0)
        x = cp.Variable(pos=True, bounds=[p, p])
        cp.Problem(cp.Minimize(x)).solve(SOLVER, gp=True)
        self.assertAlmostEqual(x.value, 3.0, places=4)

    def test_solve_without_param_value_raises(self) -> None:
        p = cp.Parameter(pos=True)
        x = cp.Variable(pos=True, bounds=[p, None])
        with pytest.raises(Exception):
            cp.Problem(cp.Minimize(x), [x <= 10.0]).solve(SOLVER, gp=True)

    def test_non_log_log_affine_bounds_rejected(self) -> None:
        """Bounds that are not log-log-affine should make the variable not DGP DPP."""
        p = cp.Parameter(3, pos=True)

        # cp.norm(p) is not log-log-affine, so this should not be DPP in DGP
        x = cp.Variable(pos=True, bounds=[cp.norm(p), None])
        assert not x.is_dpp('dgp')
        assert not cp.Problem(cp.Minimize(cp.sqrt(x))).is_dpp('dgp')

        # cp.sum(p) is log-log-convex but not log-log-affine
        y = cp.Variable(pos=True, bounds=[cp.sum(p), None])
        assert not y.is_dpp('dgp')
        assert not cp.Problem(cp.Minimize(cp.sqrt(y))).is_dpp('dgp')

        # For comparison, p[0] (indexing) is log-log-affine, so this should be DPP
        z = cp.Variable(pos=True, bounds=[p[0], None])
        assert z.is_dpp('dgp')
        assert cp.Problem(cp.Minimize(cp.sqrt(z))).is_dpp('dgp')



import numpy as np
import qdldl
import scipy.linalg as la
import scipy.sparse as sp
import scipy.sparse.linalg as sparla
from scipy.sparse import csc_array

import cvxpy.settings as settings


def orth(V, tol=1e-12):
    """Return a matrix whose columns are an orthonormal basis for range(V)"""
    Q, R, p = la.qr(V, mode='economic', pivoting=True)
    # ^ V[:, p] == Q @ R.
    rank = np.count_nonzero(np.sum(np.abs(R) > tol, axis=1))
    Q = Q[:, :rank].reshape((V.shape[0], rank))  # ensure 2-dimensional
    return Q


def onb_for_orthogonal_complement(V):
    """
    Let U = the orthogonal complement of range(V).

    This function returns an array Q whose columns are
    an orthonormal basis for U. It requires that dim(U) > 0.
    """
    n = V.shape[0]
    Q1 = orth(V)
    rank = Q1.shape[1]
    assert n > rank
    if np.iscomplexobj(V):
        P = np.eye(n) - Q1 @ Q1.conj().T
    else:
        P = np.eye(n) - Q1 @ Q1.T
    Q2 = orth(P)
    return Q2


def is_diagonal(A):
    if sp.issparse(A):
        off_diagonal_elements = A - sp.diags_array(A.diagonal())
        off_diagonal_elements = off_diagonal_elements.toarray()
    elif isinstance(A, np.ndarray):
        off_diagonal_elements = A - np.diag(np.diag(A))
    else:
        raise ValueError("Unsupported matrix type.")

    return np.allclose(off_diagonal_elements, 0)


def is_psd_within_tol(A, tol):
    """
    Return True if we can certify that A is PSD (up to tolerance "tol").

    First we check if A is PSD according to the Gershgorin Circle Theorem.

    If Gershgorin is inconclusive, then we use an iterative method (from ARPACK,
    as called through SciPy) to estimate extremal eigenvalues of certain shifted
    versions of A. The shifts are chosen so that the signs of those eigenvalues
    tell us the signs of the eigenvalues of A.

    If there are numerical issues then it's possible that this function returns
    False even when A is PSD. If you know that you're in that situation, then
    you should replace A by

        A = cvxpy.atoms.affine.wraps.psd_wrap(A).

    Parameters
    ----------
    A : Union[np.ndarray, sp.sparray]
        Symmetric (or Hermitian) NumPy ndarray or SciPy sparse array.

    tol : float
        Nonnegative. Something very small, like 1e-10.
    """

    if gershgorin_psd_check(A, tol):
        return True

    if is_diagonal(A):
        if isinstance(A, csc_array):
            return np.all(A.data >= -tol)
        else:
            min_diag_entry = np.min(np.diag(A))
            return min_diag_entry >= -tol

    def SA_eigsh(sigma):

        # Check for default_rng in np.random module (new API)
        if hasattr(np.random, 'default_rng'):
            g = np.random.default_rng(123)
        else:  # fallback to legacy RandomState
            g = np.random.RandomState(123)

        n = A.shape[0]
        v0 = g.normal(loc=0.0, scale=1.0, size=n)

        return sparla.eigsh(A, k=1, sigma=sigma, which='SA', v0=v0,
                            return_eigenvectors=False)
        # Returns the eigenvalue w[i] of A where 1/(w[i] - sigma) is minimized.
        #
        # If A - sigma*I is PSD, then w[i] should be equal to the largest
        # eigenvalue of A.
        #
        # If A - sigma*I is not PSD, then w[i] should be the largest eigenvalue
        # of A where w[i] - sigma < 0.
        #
        # We should only call this function with sigma < 0. In this case, if
        # A - sigma*I is not PSD then A is not PSD, and w[i] < -abs(sigma) is
        # a negative eigenvalue of A. If A - sigma*I is PSD, then we obviously
        # have that the smallest eigenvalue of A is >= sigma.

    try:
        ev = SA_eigsh(-tol)  # might return np.NaN, or raise exception
    except sparla.ArpackNoConvergence as e:
        # This is a numerical issue. We can't certify that A is PSD.

        message = """
        CVXPY note: This failure was encountered while trying to certify
        that a matrix is positive semi-definite (see [1] for a definition).
        In rare cases, this method fails for numerical reasons even when the matrix is
        positive semi-definite. If you know that you're in that situation, you can
        replace the matrix A by cvxpy.psd_wrap(A).

        [1] https://en.wikipedia.org/wiki/Definite_matrix
        """

        error_with_note = f"{str(e)}\n\n{message}"

        raise sparla.ArpackNoConvergence(error_with_note, e.eigenvalues, e.eigenvectors)

    if np.isnan(ev).any():
        # will be NaN if A has an eigenvalue which is exactly -tol
        # (We might also hit this code block for other reasons.)
        temp = tol - np.finfo(A.dtype).eps
        ev = SA_eigsh(-temp)

    return np.all(ev >= -tol)


def gershgorin_psd_check(A, tol):
    """
    Use the Gershgorin Circle Theorem

        https://en.wikipedia.org/wiki/Gershgorin_circle_theorem

    As a sufficient condition for A being PSD with tolerance "tol".

    The computational complexity of this function is O(nnz(A)).

    Parameters
    ----------
    A : Union[np.ndarray, sp.sparray]
        Symmetric (or Hermitian) NumPy ndarray or SciPy sparse array.

    tol : float
        Nonnegative. Something very small, like 1e-10.

    Returns
    -------
    True if A is PSD according to the Gershgorin Circle Theorem.
    Otherwise, return False.
    """
    if sp.issparse(A):
        diag = A.diagonal()
        if np.any(diag < -tol):
            return False
        A_shift = A - sp.diags_array(diag)
        A_shift = np.abs(A_shift)
        radii = np.array(A_shift.sum(axis=0)).ravel()
        return np.all(diag - radii >= -tol)
    elif isinstance(A, np.ndarray):
        diag = np.diag(A)
        if np.any(diag < -tol):
            return False
        A_shift = A - np.diag(diag)
        A_shift = np.abs(A_shift)
        radii = A_shift.sum(axis=0)
        return np.all(diag - radii >= -tol)
    else:
        raise ValueError()


class SparseCholeskyMessages:

    ASYMMETRIC = 'Input matrix is not symmetric to within provided tolerance.'
    INDEFINITE = 'Input matrix is neither positive nor negative definite.'
    FACTORIZATION_FAILED = 'Cholesky factorization failed.'
    NOT_CONST = 'The only allowed Expression inputs are Constant objects.'
    NOT_SPARSE = 'Non-Expression inputs must be SciPy sparse matrices.'
    NOT_REAL = 'Input matrix must be real.'


def dense_ldl_decomp(P, lower: bool = True, check_finite: bool = True):
    """Bunch-Kaufman LDL of a dense symmetric matrix, with 2x2 blocks resolved.

    Returns ``(diag_vals, lu)`` such that ``P = lu @ diag(diag_vals) @ lu.T``,
    where ``lu`` is the row-permuted (unit) triangular factor returned by
    ``scipy.linalg.ldl`` and any 2x2 blocks in ``D`` have been diagonalized
    via ``eigh`` (with the corresponding pair of columns of ``lu`` rotated
    to absorb the block's eigenvectors).  Used as a shared building block
    by :func:`sparse_cholesky` (when its QDLDL path fails) and
    :func:`cvxpy.atoms.quad_form.decomp_quad`.
    """
    lu, d, _perm = la.ldl(P, lower=lower, check_finite=check_finite)
    sub_diag = np.diag(d, -1)
    block_starts = np.nonzero(sub_diag)[0]
    diag_vals = np.real(np.diag(d)).copy()
    if block_starts.size:
        bs = block_starts
        idx = bs[:, None] + np.arange(2)[None, :]
        blocks = d[idx[:, :, None], idx[:, None, :]]
        eigvals, eigvecs = np.linalg.eigh(blocks)
        diag_vals[bs] = eigvals[:, 0]
        diag_vals[bs + 1] = eigvals[:, 1]
        lu_i = lu[:, bs].copy()
        lu_ip1 = lu[:, bs + 1].copy()
        lu[:, bs] = lu_i * eigvecs[:, 0, 0] + lu_ip1 * eigvecs[:, 1, 0]
        lu[:, bs + 1] = lu_i * eigvecs[:, 0, 1] + lu_ip1 * eigvecs[:, 1, 1]
    return diag_vals, lu


def _dense_ldl_factor(A, tol):
    """Fallback factor via dense pivoted LDL (Bunch-Kaufman, ``scipy.linalg.ldl``).

    Used when QDLDL cannot factor the matrix — either because it throws
    "not quasi-definite" (singular leading minors), or because it returns
    near-zero pivots paired with huge L entries whose dropped contribution
    is non-negligible.  Returns ``(sign, L, p)`` matching the
    ``sparse_cholesky`` contract.
    """
    A_dense = A.toarray() if sp.issparse(A) else np.asarray(A)
    n = A_dense.shape[0]
    diag_vals, lu = dense_ldl_decomp(A_dense, lower=True, check_finite=False)
    scale = np.max(np.abs(diag_vals)) if diag_vals.size else 0.0
    if scale == 0:
        return 1.0, sp.csr_array((n, 0)), np.arange(n)
    d_rel = diag_vals / scale
    pos = d_rel > tol
    neg = d_rel < -tol
    if pos.any() and neg.any():
        raise ValueError(SparseCholeskyMessages.INDEFINITE)
    if pos.any():
        sign = 1.0
        L = lu[:, pos] * np.sqrt(diag_vals[pos])
    else:
        sign = -1.0
        L = lu[:, neg] * np.sqrt(-diag_vals[neg])
    return sign, sp.csr_array(L), np.arange(n)


def _qdldl_residual_norm(A, sign, L, p):
    """Frobenius norm of (sign * A - L[p, :] @ L[p, :].T), used to validate
    QDLDL's factorization when zero-pivot columns were dropped."""
    Lp = L[p, :]
    resid = sign * A - (Lp @ Lp.T)
    if sp.issparse(resid):
        resid = resid.toarray()
    return np.linalg.norm(resid)


def sparse_cholesky(A, sym_tol=settings.CHOL_SYM_TOL, assume_psd=False):
    """
    The input matrix A must be real and symmetric. If A is positive (semi)definite
    then QDLDL will be used to compute its sparse LDL factorization with fill-reducing
    ordering, which is then converted to Cholesky form with zero-pivot columns dropped.
    If A is negative (semi)definite, then the analogous operation will be applied to -A.

    If factorization succeeds, then we return (sign, L, p) where sign is +1.0
    for positive (semi)definite or -1.0 for negative (semi)definite, L is an
    n-by-k matrix in CSR-format (where k = rank(A)), and p is a permutation
    vector so that (L[p, :]) @ (L[p, :]).T == sign * A within numerical precision.

    We raise a ValueError if QDLDL's factorization reveals indefiniteness or if
    we certify indefiniteness before calling QDLDL. While checking for
    indefiniteness, we also check that
     ||A - A'||_Fro / sqrt(n) <= sym_tol, where n is the order of the matrix.

    QDLDL is designed for quasi-definite matrices; on rank-deficient PSD/NSD
    inputs it can either throw "not quasi-definite" or silently return tiny
    pivots paired with huge L entries whose contribution to ``L D L.T`` is
    non-negligible.  Both modes trigger a fallback to a dense pivoted LDL
    (``scipy.linalg.ldl``) factorization.
    """
    import cvxpy.expressions.cvxtypes as cvxtypes

    if isinstance(A, cvxtypes.expression()):
        if not isinstance(A, cvxtypes.constant()):
            raise ValueError(SparseCholeskyMessages.NOT_CONST)
        A = A.value

    if not sp.issparse(A):
        raise ValueError(SparseCholeskyMessages.NOT_SPARSE)
    if np.iscomplexobj(A):
        raise ValueError(SparseCholeskyMessages.NOT_REAL)

    if not assume_psd:
        # check that we're symmetric
        symdiff = A - A.T
        sz = symdiff.data.size
        if sz > 0 and la.norm(symdiff.data) > sym_tol * (sz**0.5):
            raise ValueError(SparseCholeskyMessages.ASYMMETRIC)
        # check a necessary condition for positive/negative semidefiniteness.
        d = A.diagonal()
        maybe_psd = np.all(d >= 0)
        maybe_nsd = np.all(d <= 0)
        if not (maybe_psd or maybe_nsd):
            raise ValueError(SparseCholeskyMessages.INDEFINITE)

    n = A.shape[0]

    # QDLDL cannot handle rows/columns that are entirely zero (it fails with
    # "Matrix not properly upper-triangular"). Detect zero-diagonal entries
    # and reduce to the nonzero submatrix. For a PSD/NSD matrix, zero diagonal
    # implies all-zero row/column; we validate this before reducing.
    diag = A.diagonal()
    nonzero_idx = np.flatnonzero(diag)
    if nonzero_idx.size == 0:
        # All-zero matrix: return n-by-0 factor.
        return 1.0, sp.csr_array((n, 0)), np.arange(n)
    if nonzero_idx.size < n:
        # Validate that zero-diagonal rows are truly all-zero. If not, the
        # matrix is indefinite (e.g. [[1, 2], [2, 0]]).
        zero_idx = np.setdiff1d(np.arange(n), nonzero_idx)
        if A[zero_idx, :].nnz > 0:
            raise ValueError(SparseCholeskyMessages.INDEFINITE)
        # Reduce to the nonzero submatrix, factorize, then expand back.
        A_sub = A[np.ix_(nonzero_idx, nonzero_idx)]
        sign, L_sub, p_sub = sparse_cholesky(A_sub, sym_tol, assume_psd=True)
        # Expand L_sub back to L (n-by-k) by inserting zero rows.
        k = L_sub.shape[1]
        L_expanded = sp.lil_array((n, k))
        L_expanded[nonzero_idx, :] = L_sub[p_sub, :]
        L_expanded = sp.csr_array(L_expanded)
        return sign, L_expanded, np.arange(n)

    # QDLDL expects upper triangular CSC format
    A_upper = sp.triu(A, format='csc')

    tol = settings.CHOL_ZERO_PIVOT_TOL

    try:
        solver = qdldl.Solver(A_upper, upper=True)
        L_unit, D, p = solver.factors()
    except RuntimeError:
        # QDLDL refused to factorize (typically "not quasi-definite" on
        # rank-deficient PSD/NSD inputs, or AMD / elimination-tree
        # failures).  Fall back to dense pivoted LDL.
        return _dense_ldl_factor(A, tol)

    # QDLDL returns: A[p,:][:,p] = (I + L) @ diag(D) @ (I + L).T
    # Determine sign from D: all D >= 0 means PSD, all D <= 0 means NSD.
    scale = np.max(np.abs(D))
    if scale == 0:
        # All pivots are zero — zero matrix.
        return 1.0, sp.csr_array((n, 0)), np.arange(n)

    is_psd = np.all(D >= -tol * scale)
    is_nsd = np.all(D <= tol * scale)
    if not (is_psd or is_nsd):
        raise ValueError(SparseCholeskyMessages.INDEFINITE)

    sign = 1.0 if is_psd else -1.0

    # Build L_chol = (I + L_unit) @ diag(sqrt(|D|)), keeping only columns
    # where |D| is significantly nonzero to get an n-by-rank(A) factor.
    abs_D = np.abs(D)
    mask = abs_D > tol * scale
    sqrt_D_vals = np.sqrt(np.maximum(abs_D[mask], 0))
    I_plus_L = sp.eye_array(n, format='csr') + sp.csr_array(L_unit)
    L_chol = I_plus_L[:, mask] @ sp.diags_array(sqrt_D_vals, format='csr')
    L_chol = sp.csr_array(L_chol)

    # QDLDL's permutation p satisfies A[p,:][:,p] = L_full @ L_full.T.
    # The caller expects L_out[p_out, :] @ L_out[p_out, :].T == sign * A,
    # so we need p_out = argsort(p) (the inverse permutation).
    p_inv = np.argsort(p)

    # If we dropped any columns (rank-deficient case), QDLDL may have
    # paired tiny pivots with huge L entries whose D[j]*L[:,j]L[:,j].T
    # contribution is non-negligible; the masked product then differs
    # from sign*A by O(1).  Validate, and fall back to dense LDL if so.
    if not mask.all():
        A_norm = la.norm(A.data) if A.nnz else 0.0
        residual_tol = np.sqrt(tol) * (A_norm + 1.0)
        if _qdldl_residual_norm(A, sign, L_chol, p_inv) > residual_tol:
            return _dense_ldl_factor(A, tol)

    return sign, L_chol, p_inv



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import numpy as np
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.tests.nlp_tests.derivative_checker import DerivativeChecker


@pytest.mark.skipif('IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')
class TestAffineAtoms:

    def test_reshape_C_order(self):
        """reshape with order='C' uses the transpose-based decomposition."""
        np.random.seed(0)
        m, n = 3, 4
        X = cp.Variable((m, n), bounds=[-1, 1], name='X')
        X.value = np.random.rand(m, n)
        obj = cp.sum(cp.reshape(cp.nlp.sin(X), (m * n,), order='C'))
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_vstack(self):
        """vstack of nonlinear expressions."""
        np.random.seed(0)
        n = 5
        x = cp.Variable(n, bounds=[-1, 1], name='x')
        x.value = np.random.rand(n)
        stacked = cp.vstack([cp.nlp.sin(x), cp.nlp.cos(x)])
        obj = cp.sum(stacked)
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_upper_tri(self):
        """upper_tri of a nonlinear square matrix expression."""
        np.random.seed(0)
        n = 4
        X = cp.Variable((n, n), bounds=[-1, 1], name='X')
        X.value = np.random.rand(n, n)
        obj = cp.sum(cp.upper_tri(cp.nlp.sin(X)))
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_diag_mat(self):
        """diag(X) on a square matrix: matrix -> vector diagonal."""
        np.random.seed(0)
        n = 4
        X = cp.Variable((n, n), bounds=[-1, 1], name='X')
        X.value = np.random.rand(n, n)
        obj = cp.sum(cp.diag(cp.nlp.sin(X)))
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_div_by_scalar(self):
        """Scalar DivExpression: x / c with scalar c."""
        np.random.seed(0)
        n = 5
        x = cp.Variable(n, bounds=[-1, 1], name='x')
        x.value = np.random.rand(n)
        obj = cp.sum(cp.nlp.sin(x) / 3.0)
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_div_by_vector(self):
        """Elementwise DivExpression: x / c with vector c."""
        np.random.seed(0)
        n = 5
        c = np.arange(1, n + 1, dtype=float)
        x = cp.Variable(n, bounds=[-1, 1], name='x')
        x.value = np.random.rand(n)
        obj = cp.sum(cp.nlp.sin(x) / c)
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_conv_constant_kernel(self):
        """cp.conv(kernel, f(x)) with a constant 1D kernel."""
        np.random.seed(0)
        n = 6
        kernel = np.array([0.5, -1.0, 0.25])
        x = cp.Variable(n, bounds=[-1, 1], name='x')
        x.value = np.random.rand(n)
        obj = cp.sum(cp.conv(kernel, cp.nlp.sin(x)))
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_convolve_alias(self):
        """cp.convolve routes to the same converter as cp.conv."""
        np.random.seed(0)
        n = 6
        kernel = np.array([0.3, 0.6, -0.2])
        x = cp.Variable(n, bounds=[-1, 1], name='x')
        x.value = np.random.rand(n)
        obj = cp.sum(cp.convolve(kernel, cp.nlp.sin(x)))
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_matmul_1d_dot_product(self):
        """x.T @ x for a 1D variable. Pins both the transpose 1D no-op and
        the var-var matmul handling of a 1D right operand (numpy column-vector
        convention)."""
        np.random.seed(0)
        n = 5
        x = cp.Variable(n, bounds=[-1, 1], name='x')
        x.value = np.random.rand(n)
        assert (cp.nlp.sin(x).T @ cp.nlp.sin(x)).shape == ()

        obj = cp.nlp.sin(x).T @ cp.nlp.sin(x)
        problem = cp.Problem(cp.Minimize(obj))

        problem.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        assert(problem.status == cp.OPTIMAL)

        checker = DerivativeChecker(problem)
        checker.run_and_assert()



"""Tests for parametric (expression) bounds on variables."""

import numpy as np
import pytest

import cvxpy as cp
import cvxpy.error as error
import cvxpy.reductions.eval_params

SOLVER = cp.CLARABEL
BOUNDED_SOLVERS = ["HIGHS", "DAQP"]


def _solver_available(solver_name):
    try:
        cp.Problem(cp.Minimize(0)).solve(solver=solver_name)
        return True
    except Exception:
        return False


def _skip_if_unavailable(solver_name):
    if not _solver_available(solver_name):
        pytest.skip(f"{solver_name} not available")


class TestParametricBoundsCreation:
    """Tests for creating variables with parametric bounds."""

    def test_basic_creation(self):
        lb, ub = cp.Parameter(), cp.Parameter()
        x = cp.Variable(bounds=[lb, ub])
        assert isinstance(x.bounds[0], cp.Expression)
        assert isinstance(x.bounds[1], cp.Expression)

    def test_expression_bounds(self):
        p = cp.Parameter()
        x = cp.Variable(bounds=[p + 1, 2 * p])
        assert x.bounds is not None

    def test_mixed_none_and_param(self):
        ub = cp.Parameter()
        x = cp.Variable(bounds=[None, ub])
        assert isinstance(x.bounds[0], np.ndarray)
        assert np.all(x.bounds[0] == -np.inf)
        assert isinstance(x.bounds[1], cp.Expression)

    def test_mixed_numeric_and_param(self):
        ub = cp.Parameter()
        x = cp.Variable(bounds=[0, ub])
        assert isinstance(x.bounds[0], np.ndarray)
        assert isinstance(x.bounds[1], cp.Expression)

    def test_scalar_param_broadcast(self):
        x = cp.Variable((3,), bounds=[cp.Parameter(), cp.Parameter()])
        assert x.bounds is not None

    def test_invalid_shape_raises(self):
        with pytest.raises(ValueError):
            cp.Variable((2, 2), bounds=[cp.Parameter((3,)), 10])

    def test_parameter_rejects_expression_bounds(self):
        with pytest.raises(ValueError, match="Parametric bounds"):
            cp.Parameter(bounds=[cp.Parameter(), 10])

    def test_variable_rejects_variable_dependent_bounds(self):
        y = cp.Variable()
        with pytest.raises(ValueError, match="must not depend on Variables"):
            cp.Variable(bounds=[y, 10])
        with pytest.raises(ValueError, match="must not depend on Variables"):
            cp.Variable(bounds=[0, y])
        with pytest.raises(ValueError, match="must not depend on Variables"):
            cp.Variable(bounds=[y, y + 1])
        with pytest.raises(ValueError, match="must not depend on Variables"):
            cp.Variable(bounds=[y + cp.Parameter(), 10])

    def test_has_lower_upper_bounds(self):
        assert cp.Variable(bounds=[cp.Parameter(), 10])._has_lower_bounds()
        assert cp.Variable(bounds=[0, cp.Parameter()])._has_upper_bounds()

    def test_parameters_discovered(self):
        lb, ub = cp.Parameter(name="lb"), cp.Parameter(name="ub")
        x = cp.Variable(bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))
        ids = {p.id for p in prob.parameters()}
        assert lb.id in ids and ub.id in ids


class TestParametricBoundsSolving:
    """Tests that parametric bounds work across solver paths."""

    def test_minimize_maximize(self):
        lb, ub = cp.Parameter(value=2), cp.Parameter(value=5)
        x = cp.Variable(bounds=[lb, ub])

        prob_min = cp.Problem(cp.Minimize(x))
        prob_min.solve(solver=SOLVER)
        assert np.isclose(x.value, 2, atol=1e-4)

        prob_max = cp.Problem(cp.Maximize(x))
        prob_max.solve(solver=SOLVER)
        assert np.isclose(x.value, 5, atol=1e-4)

    def test_re_solve(self):
        lb, ub = cp.Parameter(value=2), cp.Parameter(value=5)
        x = cp.Variable(bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, 2, atol=1e-4)

        lb.value = 3
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, 3, atol=1e-4)

    def test_expression_bounds(self):
        scale = cp.Parameter(nonneg=True, value=10)
        x = cp.Variable(bounds=[-scale, scale])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, -10, atol=1e-4)

    def test_mixed_numeric_and_param(self):
        ub = cp.Parameter(value=5)
        x = cp.Variable(bounds=[0, ub])
        prob = cp.Problem(cp.Maximize(x))
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, 5, atol=1e-4)

    def test_one_sided_param_bound(self):
        ub = cp.Parameter(value=3)
        x = cp.Variable(bounds=[None, ub])
        prob = cp.Problem(cp.Maximize(x))
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, 3, atol=1e-4)

    def test_vector_variable_scalar_param(self):
        lb = cp.Parameter(value=-2)
        x = cp.Variable(3, bounds=[lb, None])
        prob = cp.Problem(cp.Minimize(cp.sum(x)), [x <= 10])
        prob.solve(solver=SOLVER)
        assert np.allclose(x.value, -2 * np.ones(3), atol=1e-4)

    def test_multiple_variables(self):
        lb1, lb2 = cp.Parameter(value=1), cp.Parameter(value=2)
        x = cp.Variable(bounds=[lb1, None])
        y = cp.Variable(bounds=[lb2, None])
        prob = cp.Problem(cp.Minimize(x + y))
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, 1, atol=1e-4)
        assert np.isclose(y.value, 2, atol=1e-4)

    def test_matrix_variable(self):
        lb = cp.Parameter((2, 3), value=np.ones((2, 3)))
        ub = cp.Parameter((2, 3), value=5 * np.ones((2, 3)))
        X = cp.Variable((2, 3), bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(cp.sum(X)))
        prob.solve(solver=SOLVER)
        np.testing.assert_allclose(X.value, np.ones((2, 3)), atol=1e-4)

    @pytest.mark.parametrize("solver_name", BOUNDED_SOLVERS)
    def test_bounded_solver_basic(self, solver_name):
        _skip_if_unavailable(solver_name)
        lb, ub = cp.Parameter(value=2), cp.Parameter(value=5)
        x = cp.Variable(bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=solver_name)
        assert np.isclose(x.value, 2, atol=1e-6)

    @pytest.mark.parametrize("solver_name", BOUNDED_SOLVERS)
    def test_bounded_solver_re_solve(self, solver_name):
        _skip_if_unavailable(solver_name)
        lb, ub = cp.Parameter(value=1), cp.Parameter(value=5)
        x = cp.Variable(bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=solver_name)
        assert np.isclose(x.value, 1, atol=1e-6)

        lb.value = 3
        prob.solve(solver=solver_name)
        assert np.isclose(x.value, 3, atol=1e-6)


class TestParametricBoundsEdgeCases:
    """Edge cases for parametric bounds."""

    def test_infinite_bounds_unbounded(self):
        ub = cp.Parameter(value=5)
        x = cp.Variable(bounds=[-np.inf, ub])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=SOLVER)
        assert prob.status in (cp.UNBOUNDED, cp.UNBOUNDED_INACCURATE)

    def test_equal_bounds_pins_value(self):
        b = cp.Parameter(value=7)
        x = cp.Variable(bounds=[b, b])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=SOLVER)
        assert np.isclose(x.value, 7, atol=1e-4)

    def test_infeasible_bounds(self):
        lb, ub = cp.Parameter(value=10), cp.Parameter(value=5)
        x = cp.Variable(bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=SOLVER)
        assert prob.status in (cp.INFEASIBLE, cp.INFEASIBLE_INACCURATE)

    def test_unset_parameter_raises(self):
        x = cp.Variable(bounds=[cp.Parameter(), 10])
        prob = cp.Problem(cp.Minimize(x))
        with pytest.raises(Exception):
            prob.solve(solver=SOLVER)


class TestParametricBoundsDPP:
    """DPP tests for parametric bounds."""

    def _make_prob(self):
        lb, ub = cp.Parameter(), cp.Parameter()
        x = cp.Variable(bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(x))
        return lb, ub, x, prob

    def test_is_dpp_and_chain(self):
        _, _, _, prob = self._make_prob()
        assert prob.is_dpp()
        _, chain, _ = prob.get_problem_data(SOLVER)
        reduction_types = [type(r) for r in chain.reductions]
        assert cvxpy.reductions.eval_params.EvalParams not in reduction_types

    def test_get_problem_data_without_param_values(self):
        _, _, _, prob = self._make_prob()
        data, _, _ = prob.get_problem_data(SOLVER)
        assert data is not None

    def test_enforce_dpp_re_solve(self):
        lb, ub, x, prob = self._make_prob()
        lb.value, ub.value = 1, 10
        prob.solve(solver=SOLVER, enforce_dpp=True)
        assert np.isclose(x.value, 1, atol=1e-4)

        lb.value = 5
        prob.solve(solver=SOLVER, enforce_dpp=True)
        assert np.isclose(x.value, 5, atol=1e-4)

    def test_enforce_dpp_expression_bounds(self):
        scale = cp.Parameter(nonneg=True, value=10)
        x = cp.Variable(bounds=[-scale, scale])
        prob = cp.Problem(cp.Minimize(x))
        prob.solve(solver=SOLVER, enforce_dpp=True)
        assert np.isclose(x.value, -10, atol=1e-4)

    @pytest.mark.parametrize("solver_name", BOUNDED_SOLVERS)
    def test_enforce_dpp_bounded_solver(self, solver_name):
        _skip_if_unavailable(solver_name)
        lb, ub, x, prob = self._make_prob()
        lb.value, ub.value = 1, 10
        prob.solve(solver=solver_name, enforce_dpp=True)
        assert np.isclose(x.value, 1, atol=1e-6)

        lb.value = 5
        prob.solve(solver=solver_name, enforce_dpp=True)
        assert np.isclose(x.value, 5, atol=1e-6)

    @pytest.mark.parametrize("bound_expr", [
        lambda: cp.Parameter() * cp.Parameter(),
        lambda: cp.Parameter((2, 2)) @ cp.Parameter(2),
    ])
    def test_non_dpp_bound_raises(self, bound_expr):
        expr = bound_expr()
        n = 2 if expr.ndim > 0 else 1
        x = cp.Variable(n, bounds=[expr, None])
        prob = cp.Problem(cp.Minimize(cp.sum(x)), [x <= 10])
        assert not prob.is_dpp()
        for p in prob.parameters():
            p.value = np.ones(p.shape) if p.ndim > 0 else 1
        with pytest.raises(error.DPPError):
            prob.solve(solver=SOLVER, enforce_dpp=True)



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum

import numpy as np
import scipy.sparse as sp


class TriangleKind(Enum):
    LOWER = "lower"
    UPPER = "upper"


def tri_to_full(tri_vec: np.ndarray, n: int, triangle: TriangleKind,
                sqrt2_scaling: bool) -> np.ndarray:
    """Expand vectorized triangle(s) to full symmetric matrix/matrices.

    Parameters
    ----------
    tri_vec : numpy.ndarray
        1-D array of triangular entries.  May contain a single triangle
        of length ``n*(n+1)//2``, or ``num`` contiguous triangles of total
        length ``num * n*(n+1)//2``.
    n : int
        The matrix side length.
    triangle : TriangleKind
        Which triangle ``tri_vec`` represents.
    sqrt2_scaling : bool
        If True, divide off-diagonal entries by sqrt(2).

    Returns
    -------
    numpy.ndarray
        If a single triangle was given, a shape ``(n, n)`` symmetric matrix.
        If multiple triangles were given, shape ``(num, n, n)``.

    Notes
    -----
    Solvers vectorize triangles in **column-major** (Fortran) order, but
    ``np.tril_indices`` / ``np.triu_indices`` enumerate in **row-major** (C)
    order. Column-major lower-triangle order is the same as row-major
    upper-triangle order (and vice versa), because enumerating the lower
    triangle column-by-column produces (row, col) pairs that, when swapped,
    give the upper triangle row-by-row. So we use ``np.triu_indices`` for
    LOWER and ``np.tril_indices`` for UPPER. Since the matrix is symmetrized
    afterward, placing entries in the transposed triangle is correct.
    """
    tri_dim = n * (n + 1) // 2
    num = tri_vec.size // tri_dim
    full = np.zeros((num, n, n))
    tri_vecs = tri_vec.reshape(num, tri_dim)
    # Use the opposite triangle: see Notes above for why.
    tri_idx = np.triu_indices(n) if triangle == TriangleKind.LOWER else np.tril_indices(n)
    full[:, *tri_idx] = tri_vecs
    full += np.swapaxes(full, -2, -1)
    diag_idx = np.diag_indices(n)
    full[:, *diag_idx] /= 2
    if sqrt2_scaling:
        lo = np.tril_indices(n, k=-1)
        hi = np.triu_indices(n, k=1)
        full[:, *lo] /= np.sqrt(2)
        full[:, *hi] /= np.sqrt(2)
    if num == 1:
        return full[0]
    return full


def psd_format_mat(constr, triangle: TriangleKind,
                   sqrt2_scaling: bool) -> sp.csc_array:
    """Return a sparse matrix that extracts the svec of each PSD matrix.

    Given a PSD constraint whose argument has shape ``(*batch, n, n)``,
    build a sparse matrix ``M`` such that ``M @ vec(X, order='F')``
    produces the scaled vectorized triangle of each matrix in the batch.

    For batched constraints the F-order vec interleaves batch elements,
    so a permutation is included to de-interleave before the per-block
    triangle extraction.

    Parameters
    ----------
    constr : PSD
        The PSD constraint.
    triangle : TriangleKind
        Which triangle the solver expects.
    sqrt2_scaling : bool
        Whether off-diagonal entries are scaled by sqrt(2).
    """
    n = constr.args[0].shape[-1]
    entries = n * (n + 1) // 2

    # --- Build the single-block format matrix (entries × n²) ---
    row_arr = np.arange(entries)
    if triangle == TriangleKind.LOWER:
        tri_idx = np.tril_indices(n)
    else:
        tri_idx = np.triu_indices(n)
    col_arr = np.sort(np.ravel_multi_index(tri_idx, (n, n), order='F'))

    val_arr = np.zeros((n, n))
    val_arr[tri_idx] = np.sqrt(2) if sqrt2_scaling else 1.0
    np.fill_diagonal(val_arr, 1.0)
    val_arr = val_arr.ravel(order='F')
    val_arr = val_arr[np.nonzero(val_arr)]

    scaled_tri = sp.csc_array((val_arr, (row_arr, col_arr)), (entries, n * n))

    idx = np.arange(n * n)
    K = idx.reshape((n, n))
    row_symm = np.append(idx, K.ravel(order='F'))
    col_symm = np.append(idx, K.T.ravel(order='F'))
    val_symm = 0.5 * np.ones(2 * n * n)
    symm_matrix = sp.csc_array((val_symm, (row_symm, col_symm)))

    single_block = scaled_tri @ symm_matrix

    # --- Handle batched constraints ---
    num = constr.num_cones()
    if num == 1:
        return single_block

    block_format = sp.block_diag([single_block] * num, format='csc')

    # F-order vec of (*batch, n, n) interleaves batch elements.
    # Permute from interleaved to contiguous per-batch blocks.
    nn = n * n
    total = num * nn
    perm = np.arange(total).reshape(nn, num).T.ravel()
    perm_mat = sp.csc_array(
        (np.ones(total), (perm, np.arange(total))),
        shape=(total, total))

    return block_format @ perm_mat.T



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np

from cvxpy.expressions.constants.constant import Constant
from cvxpy.reductions.reduction import Reduction


def replace_zero_sized(expr):
    """Recursively replace zero-sized expressions with zero constants."""
    if expr.size == 0:
        return Constant(np.zeros(expr.shape))
    if not expr.args:
        return expr
    new_args = []
    changed = False
    for arg in expr.args:
        new_arg = replace_zero_sized(arg)
        if new_arg is not arg:
            changed = True
        new_args.append(new_arg)
    if not changed:
        return expr
    return expr.copy(new_args)


class EliminateZeroSized(Reduction):
    """Eliminates zero-sized expressions from a problem.

    Zero-sized variables (those with a 0 in their shape) have no elements
    and are vacuous. This reduction:
    - Drops constraints that are entirely zero-sized (vacuously true).
    - Replaces zero-sized variables with zero constants so downstream
      reductions never see them.
    - Records eliminated variables so their values can be filled in
      during solution inversion.
    """

    def accepts(self, problem) -> bool:
        return True

    def apply(self, problem):
        from cvxpy.problems.problem import Problem

        zero_vars = {v.id: v for v in problem.variables() if v.size == 0}

        # Walk each constraint's expression tree, replacing zero-sized
        # sub-expressions with zero constants and dropping zero-sized
        # constraints.
        constraints = []
        dropped_constraints = {}
        cons_id_map = {}
        any_changed = False
        for c in problem.constraints:
            if c.size == 0:
                any_changed = True
                dropped_constraints[c.id] = c
                continue
            new_args = []
            c_changed = False
            for arg in c.args:
                new_arg = replace_zero_sized(arg)
                if new_arg is not arg:
                    c_changed = True
                new_args.append(new_arg)
            if c_changed:
                any_changed = True
                data = c.get_data()
                if data is not None:
                    new_c = type(c)(*(new_args + data))
                else:
                    new_c = type(c)(*new_args)
                constraints.append(new_c)
                cons_id_map[c.id] = new_c.id
            else:
                constraints.append(c)

        # Replace zero-sized expressions in the objective.
        obj_expr = replace_zero_sized(problem.objective.expr)
        if obj_expr is not problem.objective.expr:
            any_changed = True
            objective = type(problem.objective)(obj_expr)
        else:
            objective = problem.objective

        if not any_changed and not zero_vars:
            return problem, {}

        # Find variables that were in the original problem but are
        # absent from the reduced problem.
        reduced_problem = Problem(objective, constraints)
        new_var_ids = {v.id for v in reduced_problem.variables()}
        eliminated_vars = {}
        for v in problem.variables():
            if v.id in new_var_ids:
                continue
            if v.size > 0 and v.domain:
                # Non-zero-sized variables with domain constraints
                # (bounds, nonneg, etc.) need those constraints
                # enforced. Add them back so downstream reductions
                # (and solvers that handle bounds natively) see them.
                constraints = constraints + v.domain
            else:
                # Zero-sized variables and unconstrained orphaned
                # variables are truly eliminated.
                eliminated_vars[v.id] = v

        new_problem = Problem(objective, constraints)
        self._cons_id_map = cons_id_map
        inverse_data = {
            'eliminated_vars': eliminated_vars,
            'dropped_constraints': dropped_constraints,
            'cons_id_map': cons_id_map,
        }
        return new_problem, inverse_data

    def invert(self, solution, inverse_data):
        if not inverse_data:
            return solution
        eliminated_vars = inverse_data.get('eliminated_vars', {})
        dropped_constraints = inverse_data.get('dropped_constraints', {})
        cons_id_map = inverse_data.get('cons_id_map', {})
        # Add default values for all eliminated variables.
        if solution.primal_vars is not None:
            for vid, var in eliminated_vars.items():
                solution.primal_vars[vid] = np.zeros(var.shape)
        # Map dual values back to original constraint IDs and add
        # zero duals for dropped zero-sized constraints.
        if solution.dual_vars is not None:
            for orig_id, new_id in cons_id_map.items():
                if new_id in solution.dual_vars:
                    solution.dual_vars[orig_id] = solution.dual_vars.pop(new_id)
            for cid, c in dropped_constraints.items():
                solution.dual_vars[cid] = np.zeros(c.shape)
        return solution



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import numpy as np
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.nlp_solving_chain import _set_nlp_initial_point


class TestProblem():
    """Tests for internal NLP functions in the DNLP extension."""

    @pytest.mark.parametrize("bounds, expected", [
        (None, 0.0),
        ([None, None], 0.0),
        ([-np.inf, np.inf], 0.0),
        ([None, np.inf], 0.0),
        ([-np.inf, None], 0.0),
        ([None, 3.5], 2.5),
        ([-np.inf, 3.5], 2.5),
        ([3.5, None], 4.5),
        ([3.5, np.inf], 4.5),
        ([3.5, 4.5], 4.0),
    ])
    def test_set_initial_point_scalar_bounds(self, bounds, expected):
        x = cp.Variable((3, ), bounds=bounds)
        prob = cp.Problem(cp.Minimize(cp.sum(x)))
        _set_nlp_initial_point(prob)
        assert (x.value == expected * np.ones((3, ))).all()

    def test_set_initial_point_mixed_inf_and_finite(self):
        lb = np.array([-np.inf, 3.5, -np.inf, -1.5, 2, 2.5])
        ub = np.array([-4, 4.5, np.inf, 4.5, np.inf, 4.5])
        x = cp.Variable((6, ), bounds=[lb, ub])
        prob = cp.Problem(cp.Minimize(cp.sum(x)))
        _set_nlp_initial_point(prob)
        expected = np.array([-5, 4.0, 0.0, 1.5, 3, 3.5])
        assert (x.value == expected).all()

    def test_set_initial_point_two_variables(self):
        x = cp.Variable((2, ), bounds=[-np.inf, np.inf])
        y = cp.Variable((2, ), bounds=[-3, np.inf])
        prob = cp.Problem(cp.Minimize(cp.sum(x) + cp.sum(y)))
        _set_nlp_initial_point(prob)
        assert (x.value == np.zeros((2, ))).all()
        assert (y.value == -2 * np.ones((2, ))).all()

    def test_set_initial_point_nonnegative_attributes(self):
        x = cp.Variable((2, ), nonneg=True)
        prob = cp.Problem(cp.Minimize(cp.sum(x)))
        _set_nlp_initial_point(prob)
        assert (x.value == np.ones((2, ))).all()

    def test_set_initial_point_nonnegative_attributes_and_bounds(self):
        x = cp.Variable((2, ), nonneg=True, bounds=[1, None])
        prob = cp.Problem(cp.Minimize(cp.sum(x)))
        _set_nlp_initial_point(prob)
        assert (x.value == 2 * np.ones((2, ))).all()



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from scipy import sparse

from cvxpy.reductions.solvers.nlp_solvers.nlp_solver import Bounds


class DerivativeChecker:
    """
    A utility class to verify derivative computations by comparing
    C-based diff engine results against Python-based evaluations.
    """

    def __init__(self, problem):
        """
        Initialize the derivative checker with a CVXPY problem.

        Parameters
        ----------
        problem : cvxpy.Problem
            The CVXPY problem to check derivatives for.
        """
        from cvxpy.reductions.dnlp2smooth.dnlp2smooth import Dnlp2Smooth
        from cvxpy.reductions.solvers.nlp_solvers.diff_engine import C_problem

        self.original_problem = problem
        self._coo_initialized = False

        # Apply Dnlp2Smooth to get canonicalized problem
        canon = Dnlp2Smooth().apply(problem)
        self.canonicalized_problem = canon[0]

        # Construct the C version
        print("Constructing C diff engine problem for derivative checking...")
        self.c_problem = C_problem(self.canonicalized_problem)
        print("Done constructing C diff engine problem.")

        # Construct initial point using Bounds functionality
        self.bounds = Bounds(self.canonicalized_problem)
        self.x0 = self.bounds.x0

        # Initialize constraint bounds for checking
        self.cl = self.bounds.cl
        self.cu = self.bounds.cu

    def _init_coo(self):
        if self._coo_initialized:
            return
        self.c_problem.init_jacobian_coo()
        self.c_problem.init_hessian_coo_lower_tri()
        self.jac_rows, self.jac_cols = self.c_problem.get_jacobian_sparsity_coo()
        self.hess_rows, self.hess_cols = self.c_problem.get_problem_hessian_sparsity_coo()
        self._coo_initialized = True

    def check_constraint_values(self, x=None):
        if x is None:
            x = self.x0

        # Evaluate constraints using C implementation
        c_values = self.c_problem.constraint_forward(x)

        # Evaluate constraints using Python implementation
        # First, set variable values
        x_offset = 0
        for var in self.canonicalized_problem.variables():
            var_size = var.size
            var.value = x[x_offset:x_offset + var_size].reshape(var.shape, order='F')
            x_offset += var_size

        # Now evaluate each constraint
        python_values = []
        for constr in self.canonicalized_problem.constraints:
            constr_val = constr.expr.value.flatten(order='F')
            python_values.append(constr_val)

        python_values = np.hstack(python_values) if python_values else np.array([])
        match = np.allclose(c_values, python_values, rtol=1e-10, atol=1e-10)
        return match

    def check_jacobian(self, x=None, epsilon=1e-8):
        if x is None:
            x = self.x0

        # Get Jacobian from C implementation
        self._init_coo()
        self.c_problem.constraint_forward(x)
        jac_vals = self.c_problem.eval_jacobian_vals()
        n_constraints = len(self.cl)
        n_vars = len(x)
        c_jac_dense = np.zeros((n_constraints, n_vars))
        c_jac_dense[self.jac_rows, self.jac_cols] = jac_vals

        # Compute numerical Jacobian using central differences
        n_vars = len(x)
        n_constraints = len(self.cl)
        numerical_jac = np.zeros((n_constraints, n_vars))

        # Define constraint function for finite differences
        def constraint_func(x_eval):
            return self.c_problem.constraint_forward(x_eval)

        # Compute each column using central differences
        for j in range(n_vars):
            x_plus = x.copy()
            x_minus = x.copy()
            x_plus[j] += epsilon
            x_minus[j] -= epsilon

            c_plus = constraint_func(x_plus)
            c_minus = constraint_func(x_minus)

            numerical_jac[:, j] = (c_plus - c_minus) / (2 * epsilon)

        match = np.allclose(c_jac_dense, numerical_jac, rtol=1e-4, atol=1e-5)
        return match

    def check_hessian(self, x=None, duals=None, obj_factor=1.0, epsilon=1e-8):
        if x is None:
            x = self.x0

        if duals is None:
            duals = np.random.rand(len(self.cl))

        # must run gradient because for logistic it fills some values
        self._init_coo()
        self.c_problem.gradient()
        c_hess_vals = self.c_problem.eval_hessian_vals_coo_lower_tri(obj_factor, duals)
        n_vars = len(x)
        c_hess_dense = np.zeros((n_vars, n_vars))
        c_hess_dense[self.hess_rows, self.hess_cols] = c_hess_vals
        # Symmetrize: fill upper triangle from lower
        mask = self.hess_rows != self.hess_cols
        c_hess_dense[self.hess_cols[mask], self.hess_rows[mask]] = c_hess_vals[mask]

        # Compute numerical Hessian using finite differences of the Lagrangian gradient
        # Lagrangian gradient: ∇L = obj_factor * ∇f + J^T * duals
        def lagrangian_gradient(x_eval):
            self.c_problem.objective_forward(x_eval)
            grad_f = self.c_problem.gradient()

            self.c_problem.constraint_forward(x_eval)
            jac_vals = self.c_problem.eval_jacobian_vals()
            jac = sparse.coo_matrix(
                (jac_vals, (self.jac_rows, self.jac_cols)),
                shape=(len(self.cl), len(x))
            )

            # Lagrangian gradient = obj_factor * grad_f + J^T * duals
            return obj_factor * grad_f + jac.T @ duals

        # Compute Hessian via central differences of gradient
        numerical_hess = np.zeros((n_vars, n_vars))
        for j in range(n_vars):
            x_plus = x.copy()
            x_minus = x.copy()
            x_plus[j] += epsilon
            x_minus[j] -= epsilon

            grad_plus = lagrangian_gradient(x_plus)
            grad_minus = lagrangian_gradient(x_minus)

            numerical_hess[:, j] = (grad_plus - grad_minus) / (2 * epsilon)

        # Symmetrize the numerical Hessian (average with transpose to reduce numerical errors)
        numerical_hess = (numerical_hess + numerical_hess.T) / 2

        match = np.allclose(c_hess_dense, numerical_hess, rtol=1e-4, atol=1e-6)
        return match

    def check_objective_value(self, x=None):
        """ Compare objective value from C implementation with Python implementation. """
        if x is None:
            x = self.x0

        # Evaluate objective using C implementation
        c_obj_value = self.c_problem.objective_forward(x)

        # Evaluate objective using Python implementation
        x_offset = 0
        for var in self.canonicalized_problem.variables():
            var_size = var.size
            var.value = x[x_offset:x_offset + var_size].reshape(var.shape, order='F')
            x_offset += var_size

        python_obj_value = self.canonicalized_problem.objective.expr.value

        # Compare results
        match = np.allclose(c_obj_value, python_obj_value, rtol=1e-10, atol=1e-10)

        return match

    def check_gradient(self, x=None, epsilon=1e-8):
        """ Compare C-based gradient with numerical approximation using finite differences. """
        if x is None:
            x = self.x0
        # Get gradient from C implementation
        self.c_problem.objective_forward(x)
        c_grad = self.c_problem.gradient()

        # Compute numerical gradient using central differences
        n_vars = len(x)
        numerical_grad = np.zeros(n_vars)

        def objective_func(x_eval):
            return self.c_problem.objective_forward(x_eval)

        # Compute each component using central differences
        for j in range(n_vars):
            x_plus = x.copy()
            x_minus = x.copy()
            x_plus[j] += epsilon
            x_minus[j] -= epsilon

            f_plus = objective_func(x_plus)
            f_minus = objective_func(x_minus)

            numerical_grad[j] = (f_plus - f_minus) / (2 * epsilon)

        match = np.allclose(c_grad, numerical_grad, rtol=5 * 1e-3, atol=1e-5)
        assert(match)
        return match

    def run(self, x=None):
        """ Run all derivative checks (constraints, Jacobian, and Hessian). """

        self._init_coo()
        objective_result = self.check_objective_value(x)
        gradient_result = self.check_gradient(x)
        constraints_result = self.check_constraint_values()
        jacobian_result = self.check_jacobian(x)
        hessian_result = self.check_hessian(x)

        result = {'objective': objective_result,
                  'gradient': gradient_result,
                  'constraints': constraints_result,
                  'jacobian': jacobian_result,
                  'hessian': hessian_result}

        return result

    def run_and_assert(self, x=None):
        """ Run all derivative checks and assert correctness. """
        results = self.run(x)
        for key, passed in results.items():
            assert passed, f"Derivative check failed for {key}."




# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.9.0'
version = '1.9.0'
full_version = '1.9.0'
git_revision = '261539f'
commit_count = '265'
release = True
if not release:
    version = full_version



"""
Copyright 2013 Steven Diamond

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import numpy as np
import scipy as scipy
import scipy.sparse as sp

import cvxpy.utilities as u
from cvxpy.atoms.atom import Atom
from cvxpy.atoms.axis_atom import AxisAtom, normalize_axis
from cvxpy.constraints.constraint import Constraint
from cvxpy.expressions.constants.parameter import is_param_free


class quad_over_lin(AxisAtom):
    """:math:`(sum_{ij}X^2_{ij})/y`

    When axis is specified, computes the sum of squares along that axis,
    returning a vector instead of a scalar.
    """
    _allow_complex = True

    def __init__(
        self,
        x,
        y,
        axis: int | tuple[int, ...] | None = None,
        keepdims: bool = False
    ) -> None:
        self.axis = axis
        self.keepdims = keepdims
        # Call Atom.__init__ directly since we have two args
        Atom.__init__(self, x, y)
        # Normalize axis after init so self.args is available.
        if self.axis is not None:
            ndim = len(self.args[0].shape)
            self.axis = normalize_axis(self.axis, ndim)

    @Atom.numpy_numeric
    def numeric(self, values):
        """Returns the sum of the entries of x squared over y.
        """
        x_val = values[0]
        y_val = values[1].item()
        if self.args[0].is_complex():
            squared = np.square(x_val.imag) + np.square(x_val.real)
        else:
            squared = np.square(x_val)

        return squared.sum(axis=self.axis, keepdims=self.keepdims) / y_val

    def _domain(self) -> list[Constraint]:
        """Returns constraints describing the domain of the node.
        """
        # y > 0.
        return [self.args[1] >= 0]

    def _grad(self, values):
        """Gives the (sub/super)gradient of the atom w.r.t. each argument.

        Matrix expressions are vectorized, so the gradient is a matrix.

        Args:
            values: A list of numeric values for the arguments.

        Returns:
            A list of SciPy CSC sparse matrices or None.
        """
        # Gradient not implemented for axis case
        if self.axis is not None:
            return [None, None]

        X = values[0]
        y = values[1]
        if y <= 0:
            return [None, None]
        else:
            # DX = 2X/y, Dy = -||X||^2_2/y^2
            if self.args[0].is_complex():
                Dy = -(np.square(X.real) + np.square(X.imag)).sum()/np.square(y)
            else:
                Dy = -np.square(X).sum()/np.square(y)

            # Ensure Dy is a scalar for proper sparse array construction
            Dy = float(np.asarray(Dy).item() if np.asarray(Dy).ndim > 0 else Dy)
            Dy = sp.csc_array([[Dy]])
            DX = 2.0*X/y
            # Use F-order to match CVXPY's vectorization convention
            DX = np.reshape(DX, (self.args[0].size, 1), order='F')
            DX = scipy.sparse.csc_array(DX)
            return [DX, Dy]

    def sign_from_args(self) -> tuple[bool, bool]:
        """Returns sign (is positive, is negative) of the expression.
        """
        # Always positive.
        return (True, False)

    def is_atom_convex(self) -> bool:
        """Is the atom convex?
        """
        # Disable DPP when the second argument is a parameter.
        if u.scopes.dpp_scope_active():
            return is_param_free(self.args[1])
        else:
            return True

    def is_atom_concave(self) -> bool:
        """Is the atom concave?
        """
        return False

    def is_atom_smooth(self) -> bool:
        """Is the atom smooth?"""
        return True

    def is_atom_log_log_convex(self) -> bool:
        """Is the atom log-log convex?
        """
        return True

    def is_atom_log_log_concave(self) -> bool:
        """Is the atom log-log concave?
        """
        return False

    def is_incr(self, idx) -> bool:
        """Is the composition non-decreasing in argument idx?
        """
        return (idx == 0) and self.args[idx].is_nonneg()

    def is_decr(self, idx) -> bool:
        """Is the composition non-increasing in argument idx?
        """
        return ((idx == 0) and self.args[idx].is_nonpos()) or (idx == 1)

    def validate_arguments(self) -> None:
        """Check dimensions of arguments.
        """
        if not self.args[1].is_scalar():
            raise ValueError("The second argument to quad_over_lin must be a scalar.")
        if self.args[1].is_complex():
            raise ValueError("The second argument to quad_over_lin cannot be complex.")
        # AxisAtom.validate_arguments handles axis validation
        super(quad_over_lin, self).validate_arguments()

    def is_quadratic(self) -> bool:
        """Quadratic if x is affine and y is constant.
        """
        return self.args[0].is_affine() and self.args[1].is_constant()

    def has_quadratic_term(self) -> bool:
        """A quadratic term if y is constant.
        """
        return self.args[1].is_constant()

    def is_qpwa(self) -> bool:
        """Quadratic of piecewise affine if x is PWL and y is constant.
        """
        return self.args[0].is_pwl() and self.args[1].is_constant()



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import contextlib
import threading
from collections.abc import Generator

_thread_local = threading.local()


@contextlib.contextmanager
def dpp_scope() -> Generator[None, None, None]:
    """Context manager for DPP curvature analysis

    When this scope is active, parameters are affine, not constant.
    For example, if `param` is a Parameter, then

    ```
        with dpp_scope():
            print("param is constant: ", param.is_constant())
            print("param is affine: ", param.is_affine())
    ```

    would print

        param is constant: False
        param is affine: True
    """
    prev_state = getattr(_thread_local, 'dpp_scope_active', False)
    _thread_local.dpp_scope_active = True
    try:
        yield
    finally:
        _thread_local.dpp_scope_active = prev_state


def dpp_scope_active() -> bool:
    """Returns True if a `dpp_scope` is active. """
    return getattr(_thread_local, 'dpp_scope_active', False)


@contextlib.contextmanager
def quad_form_dpp_scope() -> Generator[None, None, None]:
    """Context manager for quad_form DPP analysis with QP solvers.

    When active, QuadForm.is_atom_convex/concave() relaxes the P.is_constant()
    requirement, allowing parametric P matrices for QP solvers that can handle
    quadratic objectives directly.

    This scope is entered by the solving chain when a QP solver is selected,
    enabling DPP caching for problems like min x'Px where P is a Parameter.
    """
    prev = getattr(_thread_local, 'quad_form_dpp_scope_active', False)
    _thread_local.quad_form_dpp_scope_active = True
    try:
        yield
    finally:
        _thread_local.quad_form_dpp_scope_active = prev


def quad_form_dpp_scope_active() -> bool:
    """Returns True if a `quad_form_dpp_scope` is active."""
    return getattr(_thread_local, 'quad_form_dpp_scope_active', False)



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest

import cvxpy as cp
from cvxpy import error


class TestDNLP():
    """
    This class tests whether problems are correctly identified as DNLP
    (disciplined nonlinear programs) and whether the objective and constraints
    are correctly identified as linearizable, linearizable convex or linearizable concave.

    We adopt the convention that a function is smooth if and only if it is
    both linearizable convex and linearizable concave. This convention is analogous to DCP
    and convex programming where a function is affine iff it is both convex and concave.
    """

    def test_simple_smooth(self):
        # Define a simple nonlinear program
        x = cp.Variable()
        y = cp.Variable()
        objective = cp.Minimize(cp.log(x - 1) + cp.exp(y - 2))
        constraints = [x + y == 1]
        prob = cp.Problem(objective, constraints)

        assert prob.is_dnlp()
        assert prob.objective.expr.is_smooth()
        assert prob.constraints[0].expr.is_smooth()

    def test_abs_linearizable_convex(self):
        x = cp.Variable()
        objective = cp.Minimize(cp.abs(x))
        prob = cp.Problem(objective)
        assert objective.expr.is_linearizable_convex()
        assert prob.is_dnlp()

    def test_sqrt_linearizable_concave(self):
        x = cp.Variable()
        objective = cp.Maximize(cp.sqrt(x))
        prob = cp.Problem(objective)
        assert objective.expr.is_linearizable_concave()
        assert prob.is_dnlp()

    def test_simple_neg_expr(self):
        x = cp.Variable()
        y = cp.Variable()
        constraints = [cp.abs(x) - cp.sqrt(y) <= 5]
        assert constraints[0].is_dnlp()
        assert constraints[0].expr.is_linearizable_convex()

    def test_non_dnlp(self):
        """
        The constraint abs(x) >= 5 is smooth concave but makes
        the problem non-DNLP.
        """
        x = cp.Variable()
        constraints = [cp.abs(x) >= 5]
        # the expression is 5 + -abs(x)
        assert constraints[0].expr.is_linearizable_concave()
        assert not constraints[0].is_dnlp()

    def test_simple_composition(self):
        x = cp.Variable()
        obj1 = cp.Minimize(cp.log(cp.abs(x)))
        assert obj1.is_dnlp()

        obj2 = cp.Minimize(cp.exp(cp.norm1(x)))
        assert obj2.is_dnlp()

        expr = cp.sqrt(cp.abs(x))
        # we treat sqrt as linearizable
        assert expr.is_dnlp()

    def test_complicated_composition(self):
        x = cp.Variable()
        y = cp.Variable()
        expr = cp.minimum(cp.sqrt(cp.exp(x)), -cp.abs(y))
        assert expr.is_linearizable_concave()

        # cannot minimize a linearizable concave function
        obj = cp.Minimize(expr)
        prob = cp.Problem(obj)
        assert not prob.is_dnlp()


class TestNonDNLP:

    def test_max(self):
        x = cp.Variable(1)
        y = cp.Variable(1)

        objective = cp.Maximize(cp.maximum(x, y))

        constraints = [x - 14 == 0, y - 6 == 0]

        # assert raises DNLP error
        problem = cp.Problem(objective, constraints)
        with pytest.raises(error.DNLPError):
            problem.solve(solver=cp.IPOPT, nlp=True)

    def test_min(self):
        x = cp.Variable(1)
        y = cp.Variable(1)

        objective = cp.Minimize(cp.minimum(x, y))

        constraints = [x - 14 == 0, y - 6 == 0]

        problem = cp.Problem(objective, constraints)
        with pytest.raises(error.DNLPError):
            problem.solve(solver=cp.IPOPT, nlp=True)

    def test_max_2(self):
        # Define variables
        x = cp.Variable(3)
        y = cp.Variable(3)

        objective = cp.Maximize(cp.sum(cp.maximum(x, y)))

        constraints = [x <= 14, y <= 14]

        problem = cp.Problem(objective, constraints)
        with pytest.raises(error.DNLPError):
            problem.solve(solver=cp.IPOPT, nlp=True)




"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import numpy as np
import pytest

import cvxpy as cp
from cvxpy.reductions.solvers.defines import INSTALLED_SOLVERS
from cvxpy.tests.nlp_tests.derivative_checker import DerivativeChecker


@pytest.mark.skipif('IPOPT' not in INSTALLED_SOLVERS, reason='IPOPT is not installed.')
class TestCompositions():
    # Stress tests for affine matrix atoms in the diff engine.

    def test_left_matmul_composition(self):
        np.random.seed(0)
        X = cp.Variable((10, 10), bounds = [-0.2, 0.2])
        A = np.random.rand(10, 10)
        obj = cp.Minimize(cp.Trace(cp.exp(A @ X)))
        constraints = [X[1, 1] + X[2, 2] == 0.1]
        prob = cp.Problem(obj, constraints)
        prob.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_right_matmul_composition(self):
        np.random.seed(0)
        X = cp.Variable((10, 10), bounds = [-0.2, 0.2])
        A = np.random.rand(10, 10)
        obj = cp.Minimize(cp.sum(cp.exp(X @ A)))
        constraints = [X[1, 1] + X[2, 2] == 0.1]
        prob = cp.Problem(obj, constraints)
        prob.solve(solver=cp.IPOPT, nlp=True, verbose=False)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_multiply_linear_composition(self):
        m = 20
        n = 5
        A = np.random.rand(m, n)
        B = np.random.rand(m, n)
        X = cp.Variable((n, n), bounds = [-1, 1])
        Y = cp.Variable((n, n), bounds = [-1, 1])
        X.value = np.random.rand(n, n)
        Y.value = np.random.rand(n, n)
        obj = cp.Minimize(cp.sum(cp.multiply(A @ X, B @ Y)))
        prob = cp.Problem(obj)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(nlp=True, verbose=True)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_multiply_nonlinear_composition(self):
        m = 20
        n = 5
        A = np.random.rand(m, n)
        B = np.random.rand(m, n)
        X = cp.Variable((n, n), bounds = [-1, 1])
        Y = cp.Variable((n, n), bounds = [-1, 1])
        X.value = np.random.rand(n, n)
        Y.value = np.random.rand(n, n)
        obj = cp.Minimize(cp.sum(cp.multiply(cp.square(A @ X), cp.logistic(B @ Y))))
        prob = cp.Problem(obj)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(nlp=True, verbose=True)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_multiply_nonlinear_composition_transpose(self):
        m = 10
        n = 10
        A = np.random.rand(m, n)
        B = np.random.rand(m, n)
        X = cp.Variable((n, n), bounds = [-1, 1])
        Y = cp.Variable((n, n), bounds = [-1, 1])
        X.value = np.random.rand(n, n)
        Y.value = np.random.rand(n, n)
        obj = cp.Minimize(cp.sum(cp.multiply(cp.square((A @ X).T), cp.logistic(B @ Y))))
        prob = cp.Problem(obj)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(nlp=True, verbose=True)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_quad_form_composition(self):
        n = 25
        Q = np.random.rand(n, n)
        Q = Q + Q.T
        x = cp.Variable(n, bounds=[-1, 1])
        obj = cp.Minimize(cp.quad_form(cp.multiply(cp.nlp.sin(x), x), Q))
        prob = cp.Problem(obj)
        x.value = np.random.rand(n)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(nlp=True, verbose=True)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_quad_form_composition_two(self):
        n = 10
        Q = np.random.rand(n, n)
        Q = Q + Q.T
        x = cp.Variable(n, bounds=[-1, 1])
        y = cp.Variable(n, bounds=[-1, 1])
        obj = cp.Minimize(cp.quad_form(cp.multiply(cp.nlp.sin(x), cp.multiply(x, y)), Q))
        prob = cp.Problem(obj)
        x.value = np.random.rand(n)
        y.value = np.random.rand(n)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(solver=cp.IPOPT, nlp=True, verbose=True)
        checker = DerivativeChecker(prob)
        checker.run_and_assert()

    def test_matmul_composition_one(self):
        np.random.seed(0)
        m, n, p = 5, 7, 11
        X = cp.Variable((m, n), bounds=[-1, 1], name='X')
        Y = cp.Variable((n, p), bounds=[-2, 2], name='Y')
        Y.value = np.random.rand(n, p)
        obj = cp.sum(cp.matmul(X, cp.nlp.cos(Y)))
        problem = cp.Problem(cp.Minimize(obj))
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True)
        assert(problem.status == cp.OPTIMAL)
        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_matmul_composition_two(self):
        np.random.seed(0)
        m, n, p = 5, 5, 5
        X = cp.Variable((m, n), bounds=[-1, 1])
        Y = cp.Variable((n, p), bounds=[-2, 2])
        Y.value = np.random.rand(n, p)
        X.value = np.random.rand(m, n)
        obj = cp.sum(cp.matmul(cp.matmul(X, X), cp.nlp.cos(Y) + X))
        problem = cp.Problem(cp.Minimize(obj))
        checker = DerivativeChecker(problem)
        checker.run_and_assert()
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True)
        assert(problem.status == cp.OPTIMAL)
        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_matmul_composition_three(self):
        np.random.seed(0)
        m, n, p = 5, 5, 5
        X = cp.Variable((m, n), bounds=[-1, 1], name='X')
        Y = cp.Variable((n, p), bounds=[-2, 2], name='Y')
        Y.value = np.random.rand(n, p)
        X.value = np.random.rand(m, n)
        obj = cp.sum(cp.matmul(cp.matmul(X, X.T), (cp.nlp.cos(Y) + X).T))
        problem = cp.Problem(cp.Minimize(obj))
        checker = DerivativeChecker(problem)
        checker.run_and_assert()
        problem.solve(solver=cp.IPOPT, nlp=True, verbose=True)
        assert(problem.status == cp.OPTIMAL)
        checker = DerivativeChecker(problem)
        checker.run_and_assert()

    def test_div(self):
        n = 5
        x = cp.Variable((n, 1), bounds=[1, 2])
        A = np.random.rand(n, n)
        obj = cp.Minimize(cp.sum(1 / (x / A @ x)))
        prob = cp.Problem(obj)
        x.value = np.random.rand(n, 1) + 1
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(nlp=True, verbose=True)
        assert np.allclose(x.value, 2 * np.ones((n, 1)))

    def test_negative_power(self):
        n = 5
        x = cp.Variable((n, 1), bounds=[1, 2])
        A = np.random.rand(n, n)
        obj = cp.Minimize(cp.sum(cp.power(A @ x, -0.5)))
        prob = cp.Problem(obj)
        x.value = np.random.rand(n, 1) + 1
        checker = DerivativeChecker(prob)
        checker.run_and_assert()
        prob.solve(nlp=True, verbose=True)
        assert prob.status == cp.OPTIMAL



"""
Copyright 2025 CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from cvxpy import problems
from cvxpy.expressions.expression import Expression
from cvxpy.problems.objective import Minimize
from cvxpy.reductions.canonicalization import Canonicalization
from cvxpy.reductions.dnlp2smooth.canonicalizers import SMOOTH_CANON_METHODS as smooth_canon_methods
from cvxpy.reductions.inverse_data import InverseData


class Dnlp2Smooth(Canonicalization):
    """
    Reduce a disciplined nonlinear program to an equivalent smooth program

    This reduction takes as input (minimization) expressions and converts
    them into smooth expressions.
    """
    def __init__(self, problem=None) -> None:
        super(Canonicalization, self).__init__(problem=problem)
        self.smooth_canon_methods = smooth_canon_methods

    def accepts(self, problem):
        """A problem is always accepted"""
        return True

    def apply(self, problem):
        """Converts an expr to a smooth program"""
        inverse_data = InverseData(problem)

        inverse_data.minimize = type(problem.objective) == Minimize

        # smoothen objective function
        canon_objective, canon_constraints = self.canonicalize_tree(
            problem.objective, True)

        # smoothen constraints
        for constraint in problem.constraints:
            # canon_constr is the constraint re-expressed in terms of
            # its canonicalized arguments, and aux_constr are the constraints
            # generated while canonicalizing the arguments of the original
            # constraint
            canon_constr, aux_constr = self.canonicalize_tree(
                constraint, False)
            canon_constraints += aux_constr + [canon_constr]
            inverse_data.cons_id_map.update({constraint.id: canon_constr.id})

        new_problem = problems.problem.Problem(canon_objective,
                                               canon_constraints)
        self._cons_id_map = inverse_data.cons_id_map
        return new_problem, inverse_data

    def canonicalize_tree(self, expr, affine_above: bool) -> tuple[Expression, list]:
        """Recursively canonicalize an Expression.

        Parameters
        ----------
        expr : The expression tree to canonicalize.
        affine_above : The path up to the root node is all affine atoms.

        Returns
        -------
        A tuple of the canonicalized expression and generated constraints.
        """
        affine_atom = type(expr) not in self.smooth_canon_methods
        canon_args = []
        constrs = []
        for arg in expr.args:
            canon_arg, c = self.canonicalize_tree(arg, affine_atom and affine_above)
            canon_args += [canon_arg]
            constrs += c
        canon_expr, c = self.canonicalize_expr(expr, canon_args, affine_above)
        constrs += c
        return canon_expr, constrs

    def canonicalize_expr(self, expr, args, affine_above: bool) -> tuple[Expression, list]:
        """Canonicalize an expression, w.r.t. canonicalized arguments.

        Parameters
        ----------
        expr : The expression tree to canonicalize.
        args : The canonicalized arguments of expr.
        affine_above : The path up to the root node is all affine atoms.

        Returns
        -------
        A tuple of the canonicalized expression and generated constraints.
        """
        # Constant trees are collapsed, but parameter trees are preserved.
        if isinstance(expr, Expression) and (
                expr.is_constant() and not expr.parameters()):
            return expr, []

        if type(expr) in self.smooth_canon_methods:
            return self.smooth_canon_methods[type(expr)](expr, args)

        return expr.copy(args), []



"""
Copyright 2025, the CVXPY developers

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Atom converter registry: maps CVXPY atom names to C diff engine constructors.

Each converter receives (expr, children) where expr is the CVXPY expression
and children are already-converted C nodes. matmul and multiply are handled
separately in converters.py (they need param_dict for parameter support).
"""
import numpy as np
from scipy import sparse
from sparsediffpy import _sparsediffengine as _diffengine

from cvxpy.reductions.solvers.nlp_solvers.diff_engine.helpers import (
    chain_add,
    normalize_shape,
    to_dense_float,
)


def convert_hstack(expr, children):
    """Convert horizontal stack (hstack) of expressions."""
    return _diffengine.make_hstack(children)


def convert_vstack(expr, children):
    """Convert vertical stack (vstack) of expressions."""
    return _diffengine.make_vstack(children)


def convert_conv(expr, children):
    """Convert cp.conv / cp.convolve (full 1D convolution)."""
    return _diffengine.make_convolve(children[0], children[1])


def convert_div(expr, children):
    """Convert x / c by multiplying x by the elementwise reciprocal of c.

    Matches coo_backend.div: parametrized divisors are rejected and any
    zero entry in the divisor raises explicitly.
    """
    divisor_expr = expr.args[1]
    if divisor_expr.parameters():
        raise NotImplementedError("div doesn't support parametrized divisor")
    divisor = to_dense_float(divisor_expr.value)
    if np.any(divisor == 0):
        raise ValueError("Division by zero encountered in divisor")
    recip = 1.0 / divisor
    d1, d2 = normalize_shape(recip.shape)
    recip_node = _diffengine.make_parameter(d1, d2, -1, 0, recip.flatten(order='F'))
    if recip.size == 1:
        return _diffengine.make_param_scalar_mult(recip_node, children[0])
    return _diffengine.make_param_vector_mult(recip_node, children[0])


def extract_flat_indices_from_index(expr):
    """Extract flattened indices from CVXPY index expression."""
    parent_shape = expr.args[0].shape
    indices_per_dim = [np.arange(s.start, s.stop, s.step) for s in expr.key]

    if len(indices_per_dim) == 1:
        return indices_per_dim[0].astype(np.int32)
    elif len(indices_per_dim) == 2:
        # Fortran order: idx = row + col * n_rows
        return (
            np.add.outer(indices_per_dim[0], indices_per_dim[1] * parent_shape[0])
            .flatten(order="F")
            .astype(np.int32)
        )
    else:
        raise NotImplementedError("index with >2 dimensions not supported")


def extract_flat_indices_from_special_index(expr):
    """Extract flattened indices from CVXPY special_index expression."""
    return np.reshape(expr._select_mat, expr._select_mat.size, order="F").astype(np.int32)


def convert_rel_entr(expr, children):
    """Convert rel_entr(x, y) = x * log(x/y).

    The C engine auto-dispatches based on argument dimensions
    (elementwise, vector-scalar, or scalar-vector).
    """
    x_size = expr.args[0].size
    y_size = expr.args[1].size
    if x_size > 1 and y_size > 1 and x_size != y_size:
        raise ValueError(
            f"rel_entr requires compatible argument sizes. "
            f"Got: x={x_size}, y={y_size}"
        )
    return _diffengine.make_rel_entr(children[0], children[1])


def convert_quad_form(expr, children):
    """Convert scalar quadratic form x.T @ P @ x."""
    P = expr.args[1]

    if P.parameters():
        raise NotImplementedError(
            "quad_form with a parameterized P is not supported by the diff engine."
        )

    if not P.is_constant():
        raise NotImplementedError("quad_form requires P to be a constant matrix")

    P_val = P.value
    if P_val is None:
        raise NotImplementedError(
            "quad_form with a symbolic P (e.g. eye/parameter without a value) "
            "is not supported by the diff engine."
        )

    if not isinstance(P_val, sparse.csr_matrix):
        P_val = sparse.csr_matrix(P_val)

    return _diffengine.make_quad_form(
        children[0],
        P_val.data.astype(np.float64),
        P_val.indices.astype(np.int32),
        P_val.indptr.astype(np.int32),
        P_val.shape[0],
        P_val.shape[1],
    )


def convert_reshape(expr, children):
    """Convert reshape. C-order via transpose(F-reshape(transpose(x))).

    Identity: reshape(x, (m, n), C) == transpose(reshape(transpose(x), (n, m), F)).
    """
    d1, d2 = normalize_shape(expr.shape)
    if expr.order == "F":
        return _diffengine.make_reshape(children[0], d1, d2)
    transposed = _diffengine.make_transpose(children[0])
    reshaped = _diffengine.make_reshape(transposed, d2, d1)
    return _diffengine.make_transpose(reshaped)

def convert_broadcast(expr, children):
    d1, d2 = expr.broadcast_shape
    d1_C, d2_C = _diffengine.get_expr_dimensions(children[0])
    if d1_C == d1 and d2_C == d2:
        return children[0]

    return _diffengine.make_broadcast(children[0], d1, d2)

def convert_sum(expr, children):
    axis = expr.axis
    if axis is None:
        axis = -1
    return _diffengine.make_sum(children[0], axis)

def convert_promote(expr, children):
    d1, d2 = normalize_shape(expr.shape)
    return _diffengine.make_promote(children[0], d1, d2)

def convert_NegExpression(_expr, children):
    return _diffengine.make_neg(children[0])

def convert_quad_over_lin(_expr, children):
    return _diffengine.make_quad_over_lin(children[0], children[1])

def convert_index(expr, children):
    idxs = extract_flat_indices_from_index(expr)
    d1, d2 = normalize_shape(expr.shape)
    return _diffengine.make_index(children[0], d1, d2, idxs)

def convert_special_index(expr, children):
    idxs = extract_flat_indices_from_special_index(expr)
    d1, d2 = normalize_shape(expr.shape)
    return _diffengine.make_index(children[0], d1, d2, idxs)

def convert_prod(expr, children):
    axis = expr.axis
    if axis is None:
        return _diffengine.make_prod(children[0])
    elif axis == 0:
        return _diffengine.make_prod_axis_zero(children[0])
    elif axis == 1:
        return _diffengine.make_prod_axis_one(children[0])

def convert_transpose(expr, children):
    # If a user calls transpose on a (n, ) expression, CVXPY treats it as a
    # no-op and keeps the shape as (n, ). The diff engine represents all 1D
    # expressions as (1, n), so we need to check for this case and avoid
    # transposing if it's just a 1D expression.
    if len(expr.args[0].shape) <= 1:
        return children[0]

    # If the child is a vector (shape (n,1) or (1,n), use reshape to transpose
    child_shape = normalize_shape(expr.args[0].shape)
    if 1 in child_shape:
        return _diffengine.make_reshape(children[0], child_shape[1], child_shape[0])
    else:
        return _diffengine.make_transpose(children[0])

def convert_trace(_expr, children):
    return _diffengine.make_trace(children[0])

def convert_diag_vec(expr, children):
    # C implementation only supports k=0 (main diagonal)
    # TODO add support for diag vec with k
    if expr.k != 0:
        raise NotImplementedError("diag_vec with k != 0 not supported in diff engine")
    return _diffengine.make_diag_vec(children[0])


def convert_diag_mat(expr, children):
    """Convert diag_mat: extract diagonal from square matrix."""
    if expr.k != 0:
        raise NotImplementedError("diag_mat with k != 0 not supported in diff engine")
    node = _diffengine.make_diag_mat(children[0])
    # C produces (n, 1) but CVXPY shape is (n,) which normalizes to (1, n)
    # TODO add support for producing (1, n) directly in C and remove this reshape
    # TODO also raise error that the k should be zero, since that's the only supported case
    n = expr.args[0].shape[0]
    return _diffengine.make_reshape(node, 1, n)


def convert_upper_tri(_expr, children):
    """Convert upper_tri: extract strict upper triangular elements."""
    return _diffengine.make_upper_tri(children[0])


ATOM_CONVERTERS = {
    # Elementwise unary
    "log": lambda _expr, children: _diffengine.make_log(children[0]),
    "exp": lambda _expr, children: _diffengine.make_exp(children[0]),
    # Affine unary
    "NegExpression": convert_NegExpression,
    "Promote": convert_promote,
    # Pass-through wrappers (no-ops for real-valued expressions)
    "conj": lambda _expr, children: children[0],
    "nonneg_wrap": lambda _expr, children: children[0],
    "nonpos_wrap": lambda _expr, children: children[0],
    "psd_wrap": lambda _expr, children: children[0],
    "nsd_wrap": lambda _expr, children: children[0],
    "hermitian_wrap": lambda _expr, children: children[0],
    "skew_symmetric_wrap": lambda _expr, children: children[0],
    "symmetric_wrap": lambda _expr, children: children[0],
    # Division by constant
    "DivExpression": convert_div,
    # N-ary (handles 2+ args)
    "AddExpression": lambda _expr, children: chain_add(children),
    # Reductions
    "Sum": convert_sum,
    # Bivariate
    "QuadForm": convert_quad_form,
    "quad_over_lin": convert_quad_over_lin,
    "rel_entr": convert_rel_entr,
    # Elementwise univariate with parameter
    "Power": lambda expr, children: _diffengine.make_power(children[0], float(expr.p.value)),
    "PowerApprox": lambda expr, children: _diffengine.make_power(children[0], float(expr.p.value)),
    # Trigonometric
    "sin": lambda _expr, children: _diffengine.make_sin(children[0]),
    "cos": lambda _expr, children: _diffengine.make_cos(children[0]),
    "tan": lambda _expr, children: _diffengine.make_tan(children[0]),
    # Hyperbolic
    "sinh": lambda _expr, children: _diffengine.make_sinh(children[0]),
    "tanh": lambda _expr, children: _diffengine.make_tanh(children[0]),
    "asinh": lambda _expr, children: _diffengine.make_asinh(children[0]),
    "atanh": lambda _expr, children: _diffengine.make_atanh(children[0]),
    # Other elementwise
    "entr": lambda _expr, children: _diffengine.make_entr(children[0]),
    "logistic": lambda _expr, children: _diffengine.make_logistic(children[0]),
    "xexp": lambda _expr, children: _diffengine.make_xexp(children[0]),
    "normcdf": lambda _expr, children: _diffengine.make_normal_cdf(children[0]),
    # Indexing/slicing
    "index": convert_index,
    "special_index": convert_special_index,
    "reshape": convert_reshape,
    "broadcast_to": convert_broadcast,
    # Reductions returning scalar
    "Prod": convert_prod,
    "transpose": convert_transpose,
    # Horizontal/vertical stack
    "Hstack": convert_hstack,
    "Vstack": convert_vstack,
    # 1D full convolution
    "conv": convert_conv,
    "convolve": convert_conv,
    "Trace": convert_trace,
    # Diagonal and triangular
    "diag_vec": convert_diag_vec,
    "diag_mat": convert_diag_mat,
    "upper_tri": convert_upper_tri,
}



"""
Copyright, the CVXPY authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import pytest

import cvxpy as cp
from cvxpy.utilities import scopes

SOLVER = cp.CLARABEL
RTOL = 1e-3


def _solve(problem: cp.Problem, **kwargs) -> None:
    problem.solve(solver=SOLVER, **kwargs)
    assert problem.status == cp.OPTIMAL


def _solve_and_copy(problem: cp.Problem, variable: cp.Variable, **kwargs) -> np.ndarray:
    _solve(problem, **kwargs)
    return variable.value.copy()


def _assert_allclose(actual, expected) -> None:
    assert np.allclose(actual, expected, rtol=RTOL)


def _assert_qp_matches_conic(problem: cp.Problem, variable: cp.Variable) -> np.ndarray:
    qp_solution = _solve_and_copy(problem, variable)
    _solve(problem, use_quad_obj=False)
    _assert_allclose(qp_solution, variable.value)
    return qp_solution


def _assert_resolve_matches_fresh(
    first_solution: np.ndarray,
    resolved_solution: np.ndarray,
    fresh_problem: cp.Problem,
    variable: cp.Variable,
) -> None:
    fresh_solution = _solve_and_copy(fresh_problem, variable)
    assert not np.allclose(first_solution, resolved_solution, atol=RTOL)
    _assert_allclose(resolved_solution, fresh_solution)


class TestQuadFormDPPDetection:
    """Test that is_dpp() correctly identifies DPP-compliant problems."""

    def test_parametric_P_is_dpp_in_scope(self) -> None:
        """quad_form(x, P) with Parameter P is DPP only in quad_form_dpp_scope."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)
        y = cp.quad_form(x, P)

        assert not y.is_dpp()
        with scopes.quad_form_dpp_scope():
            assert y.is_dpp()

    def test_constant_P_always_dpp(self) -> None:
        """quad_form(x, constant) is always DPP."""
        x = cp.Variable(2)
        P = np.array([[2, 1], [1, 2]])
        assert cp.quad_form(x, P).is_dpp()

    @pytest.mark.parametrize("invalid_case", [
        "param_in_x",
        "x_plus_param",
        "param_times_param",
    ])
    def test_invalid_cases_not_dpp(self, invalid_case) -> None:
        """Cases that violate DPP requirements are correctly rejected."""
        x = cp.Variable(2)
        p = cp.Parameter(2)
        P = cp.Parameter((2, 2), PSD=True)
        P.value = np.eye(2)

        if invalid_case == "param_in_x":
            y = cp.quad_form(p, np.eye(2))
        elif invalid_case == "x_plus_param":
            y = cp.quad_form(x + p, P)
        elif invalid_case == "param_times_param":
            alpha = cp.Parameter(nonneg=True)
            alpha.value = 1.0
            y = cp.quad_form(x, alpha * P, assume_PSD=True)

        with scopes.quad_form_dpp_scope():
            assert not y.is_dpp()


class TestQuadFormDPPCompilation:
    """Test that DPP compilation works without P.value set."""

    def test_get_problem_data_without_P_value(self) -> None:
        """get_problem_data works without P.value set (DPP promise)."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)

        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P)), [cp.sum(x) == 1])
        _, chain, _ = prob.get_problem_data(solver=cp.CLARABEL)

        reduction_types = [type(r).__name__ for r in chain.reductions]
        assert 'EvalParams' not in reduction_types

    def test_is_dpp_without_P_value(self) -> None:
        """is_dpp() works without P.value set."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)

        with scopes.quad_form_dpp_scope():
            assert cp.quad_form(x, P).is_dpp()


class TestQuadFormDPPCorrectness:
    """Test that DPP path produces correct results."""

    @pytest.mark.parametrize(
        ("P_matrix", "expected_x"),
        [
            pytest.param(np.array([[2, 0], [0, 1]]), np.array([1 / 3, 2 / 3]), id="diagonal"),
            pytest.param(np.eye(2), np.array([0.5, 0.5]), id="identity"),
            pytest.param(5 * np.eye(2), np.array([0.5, 0.5]), id="scaled_identity"),
            pytest.param(
                np.array([[2, 0.5], [0.5, 1]]),
                None,
                id="dense_psd",
            ),
        ],
    )
    def test_minimize_quad_form(self, P_matrix, expected_x) -> None:
        """Minimizing a parametric quad_form matches the conic path."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)
        P.value = P_matrix

        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P)), [cp.sum(x) == 1])
        qp_solution = _assert_qp_matches_conic(prob, x)

        if expected_x is not None:
            _assert_allclose(qp_solution, expected_x)

    def test_maximize_with_nsd_P(self) -> None:
        """Maximize quad_form with NSD P produces correct solution."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), NSD=True)
        P.value = -np.eye(2)

        prob = cp.Problem(cp.Maximize(cp.quad_form(x, P)), [cp.sum(x) == 1])
        _solve(prob)

        _assert_allclose(x.value, [0.5, 0.5])
        assert np.isclose(prob.value, -0.5, rtol=RTOL)

    def test_resolve_with_different_P_values(self) -> None:
        """Re-solving with different P values works correctly."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)
        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P)), [cp.sum(x) == 1])

        P.value = np.array([[2, 0], [0, 1]])
        _solve(prob)
        _assert_allclose(x.value, [1 / 3, 2 / 3])

        P.value = np.array([[1, 0], [0, 2]])
        _solve(prob)
        _assert_allclose(x.value, [2 / 3, 1 / 3])


class TestQuadFormDPPVariants:
    """Test various problem structures with parametric quad_form."""

    def test_quad_form_plus_linear(self) -> None:
        """quad_form + linear term solves correctly."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)
        c = cp.Parameter(2)
        P.value = np.eye(2)
        c.value = np.array([2, -2])

        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P) + c @ x), [cp.sum(x) == 1])
        _solve(prob)
        _assert_allclose(x.value, [-0.5, 1.5])

    def test_quad_form_plus_constant_linear(self) -> None:
        """quad_form(x, P) + c @ x with constant c solves correctly.

        Regression test: ensures the q coefficient from the linear term is
        not overwritten when the quad_form dummy variable is processed after
        the true variable in extract_quadratic_coeffs.
        """
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)
        P.value = np.eye(2)
        c = np.array([2.0, -2.0])

        # min x'Px + c'x s.t. sum(x) == 1
        # Analytic: x* = (1 - P^{-1} c) / (1' P^{-1} 1) adjusted for constraint
        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P) + c @ x), [cp.sum(x) == 1])
        qp_solution = _assert_qp_matches_conic(prob, x)
        _assert_allclose(qp_solution, [-0.5, 1.5])

        # Re-solve with a different P to ensure DPP caching works.
        P.value = np.array([[2, 0], [0, 1]])
        _assert_qp_matches_conic(prob, x)

    def test_quad_form_in_constraint(self) -> None:
        """Parametric quad_form in constraint solves correctly."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)
        P.value = np.eye(2)

        prob = cp.Problem(cp.Minimize(cp.sum(x)), [cp.quad_form(x, P) <= 1])
        _solve(prob)
        expected = np.array([-1 / np.sqrt(2), -1 / np.sqrt(2)])
        _assert_allclose(x.value, expected)

    def test_multiple_quad_forms(self) -> None:
        """Multiple quad_forms with same variable solve correctly."""
        x = cp.Variable(2)
        P1 = cp.Parameter((2, 2), PSD=True)
        P2 = cp.Parameter((2, 2), PSD=True)
        P1.value = np.array([[2, 0], [0, 1]])
        P2.value = np.array([[1, 0], [0, 2]])

        # x'P1x + x'P2x = x'(P1+P2)x = x'[[3,0],[0,3]]x => isotropic
        prob = cp.Problem(
            cp.Minimize(cp.quad_form(x, P1) + cp.quad_form(x, P2)),
            [cp.sum(x) == 1]
        )
        _solve(prob)
        _assert_allclose(x.value, [0.5, 0.5])

    def test_affine_combination_of_params(self) -> None:
        """quad_form(x, P1 + P2) solves correctly."""
        x = cp.Variable(2)
        P1 = cp.Parameter((2, 2), PSD=True)
        P2 = cp.Parameter((2, 2), PSD=True)
        P1.value = np.array([[2, 0], [0, 0]])
        P2.value = np.array([[0, 0], [0, 1]])

        prob = cp.Problem(cp.Minimize(cp.quad_form(x, P1 + P2)), [cp.sum(x) == 1])
        _solve(prob)
        _assert_allclose(x.value, [1 / 3, 2 / 3])

    def test_resolve_scaled_param(self) -> None:
        """Re-solving quad_form(x, 2*P) updates P correctly."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)

        prob = cp.Problem(
            cp.Minimize(cp.quad_form(x, 2 * P) + 5 * x[0] + 3 * x[1])
        )

        P.value = np.array([[2, 0], [0, 1]])
        first_solution = _solve_and_copy(prob, x)

        P.value = np.array([[1, 0], [0, 5]])
        resolved_solution = _solve_and_copy(prob, x)

        # Verify against a fresh solve with the second P value.
        fresh_problem = cp.Problem(
            cp.Minimize(cp.quad_form(x, 2 * P) + 5 * x[0] + 3 * x[1])
        )
        _assert_resolve_matches_fresh(first_solution, resolved_solution, fresh_problem, x)

    def test_resolve_affine_combination_of_params(self) -> None:
        """Re-solving quad_form(x, P1 + P2) updates correctly."""
        x = cp.Variable(2)
        P1 = cp.Parameter((2, 2), PSD=True)
        P2 = cp.Parameter((2, 2), PSD=True)

        prob = cp.Problem(
            cp.Minimize(cp.quad_form(x, P1 + P2)), [cp.sum(x) == 1]
        )

        # P1+P2 = [[2,0],[0,1]] — weights x[0] more, so x[1] preferred.
        P1.value = np.array([[2, 0], [0, 0]])
        P2.value = np.array([[0, 0], [0, 1]])
        _solve(prob)
        _assert_allclose(x.value, [1 / 3, 2 / 3])

        # P1+P2 = [[1,0],[0,2]] — weights x[1] more, so x[0] preferred.
        P1.value = np.array([[1, 0], [0, 0]])
        P2.value = np.array([[0, 0], [0, 2]])
        _solve(prob)
        _assert_allclose(x.value, [2 / 3, 1 / 3])


class TestQuadFormDPPConstraintResolve:
    """Regression tests: parametric quad_form in constraints must update on re-solve.

    When quad_form(x, P) appears in a constraint (not the objective), the conic
    canonicalizer bakes in numeric Cholesky factors from the first P.value.
    If the problem is incorrectly classified as DPP, re-solving with a different
    P.value reuses stale factors and produces wrong results.
    """

    def test_constraint_quad_form_resolve(self) -> None:
        """Re-solving with different P in constraint must update the constraint."""
        x = cp.Variable(2)
        P = cp.Parameter((2, 2), PSD=True)

        prob = cp.Problem(cp.Minimize(cp.sum(x)), [cp.quad_form(x, P) <= 1])

        P.value = np.eye(2)
        first_solution = _solve_and_copy(prob, x)
        _assert_allclose(first_solution, [-1 / np.sqrt(2), -1 / np.sqrt(2)])

        # Tighter on x[0]: solution should shift toward x[1].
        P.value = np.array([[4, 0], [0, 1]])
        resolved_solution = _solve_and_copy(prob, x)

        # Cross-check against a fresh problem with the same P.
        fresh_problem = cp.Problem(cp.Minimize(cp.sum(x)), [cp.quad_form(x, P) <= 1])
        _assert_resolve_matches_fresh(first_solution, resolved_solution, fresh_problem, x)

    def test_mixed_objective_and_constraint_resolve(self) -> None:
        """Parametric P in both objective and constraint must both update."""
        x = cp.Variable(2)
        P_obj = cp.Parameter((2, 2), PSD=True)
        P_constr = cp.Parameter((2, 2), PSD=True)

        prob = cp.Problem(
            cp.Minimize(cp.quad_form(x, P_obj)),
            [cp.quad_form(x, P_constr) <= 4, cp.sum(x) == 1]
        )

        P_obj.value = np.eye(2)
        P_constr.value = np.eye(2)
        first_solution = _solve_and_copy(prob, x)

        # Change constraint P only — tighter on x[0].
        P_constr.value = np.array([[100, 0], [0, 1]])
        resolved_solution = _solve_and_copy(prob, x)

        # Cross-check against a fresh problem.
        fresh_problem = cp.Problem(
            cp.Minimize(cp.quad_form(x, P_obj)),
            [cp.quad_form(x, P_constr) <= 4, cp.sum(x) == 1]
        )
        _assert_resolve_matches_fresh(first_solution, resolved_solution, fresh_problem, x)


def test_hermitian_param_resolve() -> None:
    """Test quad_form DPP re-solve with complex hermitian P."""
    x = cp.Variable(2, complex=True)
    P = cp.Parameter((2, 2), PSD=True, complex=True)
    constrs = [cp.real(x[0]) >= 1, cp.real(x[1]) >= 1]
    prob = cp.Problem(cp.Minimize(cp.quad_form(x, P)), constrs)

    P.value = np.array([[2, 1 + 1j], [1 - 1j, 3]])
    first_solution = _solve_and_copy(prob, x)

    P.value = np.array([[4, 1 + 2j], [1 - 2j, 5]])
    resolved_solution = _solve_and_copy(prob, x)

    assert not np.allclose(first_solution, resolved_solution, atol=RTOL)
    _solve(prob, use_quad_obj=False)
    _assert_allclose(resolved_solution, x.value)
